"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import logging
import time
import serial # type: ignore

from wiliot_core import WiliotGateway, CommandDetails
from wiliot_tools.test_equipment.test_equipment import ZebraPrinter, serial_ports
from wiliot_testers.offline.modules.offline_r2r_controller import R2R as R2R_Offline


class R2RException(Exception):
    def __init__(self, msg):
        super().__init__(f'R2R: {msg}')


ARDUINO_BAUD = 1000000
ARDUINO_NAME = 'Williot R2R GPIO 1.0'


class R2RTempCalib(object):
    def __init__(self, logger_name, counter_start_idx=0):
        self.logger = logging.getLogger(logger_name)

        # connect to the r2r/arduino:
        self.r2r = None
        self.port = ''
        self.connected = False
        self.counter = int(counter_start_idx)
        self.connect()
        if not self.connected:
            raise R2RException('Could not connect to arduino!')

    def connect(self):
        baud_rate = ARDUINO_BAUD
        if self.port:
            ports_list = [self.port]
        else:
            ports_list = serial_ports()
        for p in ports_list:
            try:
                self.r2r = serial.Serial(p, baud_rate, timeout=0.1)
                time.sleep(1)
                if self.r2r.isOpen():
                    rsp = self.r2r.readline().decode()
                    if ARDUINO_NAME in rsp:
                        self.port = p
                        self.connected = True
                        self.logger.warning(f'Connect to R2R port {p}')
                        return
            except Exception as e:
                self.logger.warning(f'Could not connect to port {p} due to {e}, check the next port')

    def write(self, cmd):
        if not self.r2r.isOpen():
            self.connected = False
            self.connect()

        self.r2r.write(cmd.encode())
        time.sleep(0.5)
        rsp = self.r2r.readline()
        return rsp.decode()

    def move_r2r(self):
        rsp = self.write('T')
        self.logger.info(f'R2R responded with {rsp}')
        self.counter += 1

    def disconnect(self):
        self.connected = False
        if not self.r2r.isOpen():
            self.r2r.close()

    def get_counter(self):
        return self.counter


class R2R(object):
    def __init__(self, logger_config, counter_start_idx=0, r2r_type='arduino', r2r_printer_config={}):
        self.r2r_type = r2r_type.lower()
        if r2r_type.lower() == 'arduino':
            self.r2r = R2RTempCalib(logger_name=logger_config['logger_name'], counter_start_idx=counter_start_idx)
        elif r2r_type.lower() == 'gateway':
            self.gw = WiliotGateway(auto_connect=True, logger_name=logger_config['logger_name'],
                                    log_dir_for_multi_processes=logger_config['logger_path'], pass_packets=False)
            self.gw.set_configuration(cmds={CommandDetails.pl_gw_config: 1}, start_gw_app=False)
            self.r2r = R2R_Offline(logger_name=logger_config['logger_name'],
                                   counter_start_idx=counter_start_idx, gw_obj=self.gw)
        elif r2r_type.lower() == 'zebra printer':
            r2r_printer_config['log_name'] = logger_config['logger_name']
            self.r2r = ZebraPrinter(**r2r_printer_config)
        else:
            raise R2RException(f'Unknown r2r type: {r2r_type}')
    
    def move_r2r(self):
        if self.r2r_type == 'gateway':
            self.gw.stop_gw_app()
        return self.r2r.move_r2r()
    
    def is_r2r_move(self, timeout=0):
        if self.r2r_type == 'arduino' or self.r2r_type == 'zebra printer':
            time.sleep(timeout)
            return True
        elif self.r2r_type == 'gateway':
            rsp = self.gw.read_specific_message(msg="Production Line GW", read_timeout=timeout)
            if 'Trigger During Run' in rsp:
                raise R2RException('R2R: R2R move during test! Make sure step time is set to maximal in the machine controller')
            return rsp != ''
        return False
    
    def is_r2r_move_during_test(self):
        if self.r2r_type == 'gateway':
            rsp = self.gw.read_specific_message(msg='Trigger During Run', read_timeout=0.1)
            return rsp != ''
        return False

    def get_counter(self):
        return self.r2r.get_counter()
    
    def pause_app(self):
        if self.r2r_type == 'arduino':
            self.r2r.disconnect()
        elif self.r2r_type == 'gateway':
            self.r2r.send_stop_to_r2r()

    def continue_app(self):
        if self.r2r_type == 'arduino':
            self.r2r.connect()
        elif self.r2r_type == 'gateway':
            self.r2r.send_start_to_r2r()
    
    def exit_app(self):
        if self.r2r_type == 'arduino':
            self.r2r.disconnect()
        elif self.r2r_type == 'gateway':
            self.r2r.send_stop_to_r2r()
            time.sleep(1)
            self.gw.reset_gw()
            self.gw.exit_gw_api()


"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import json
import logging
import os
import multiprocessing
from queue import Queue
import time
import datetime
import pandas as pd
import traceback

from wiliot_core.utils.utils import QueueHandler
from wiliot_tools.association_tool.association_configs import is_asset_code, is_wiliot_code, \
    WILIOT_MIN_NUM_CODE, ASSET_NUM_CODES
from wiliot_tools.association_tool.send_association_to_cloud import CloudAssociation
from wiliot_tools.test_equipment.test_equipment import BarcodeScanner
from wiliot_testers.association_tester.hw_components.r2r_component import R2R, R2RException
from wiliot_testers.association_tester.hw_components.scanner_component import Scanner
from wiliot_testers.offline.modules.offline_printing_and_validation import Printer
pd.options.mode.chained_assignment = None  # default='warn'


ASSOCIATION_Q_SIZE = 200


class AssociatorProcess(CloudAssociation):
    def __init__(self, associate_q=None, associated_status_q=None, stop_event=None, owner_id='', category_id='', time_btwn_request=1,
                 initiator_name=None, common_run_name=''):
        self.associated_status_q = associated_status_q
        super().__init__(associate_q=associate_q, stop_event=stop_event, owner_id=owner_id,
                         is_gcp=False, category_id=category_id, time_btwn_request=time_btwn_request,
                         initiator_name=initiator_name,
                         logger_config={'dir_name': 'association_and_verification_tester',
                                        'folder_name': common_run_name})

    def handle_results(self, message, asset_id, pixel_dict, bad_association):
        bad_association = super().handle_results(message, asset_id, pixel_dict, bad_association)
        element = {'is_associated': int(message['status_code'] // 100) == 2,
                   'associate_status_code': message['status_code'],
                   'asset_code': [asset_id],
                   'wiliot_code': pixel_dict['pixel_id']}
        if self.associated_status_q.full():
            self.logger.warning(f'associated_status_q queue is full. discard {element}')
        else:
            self.associated_status_q.put(element)
        return bad_association


class TagAssociation(object):
    def __init__(self, user_input, stop_event, rescan_event, is_app_running, logger_config, exception_q, run_param_file):
        self.logger = logging.getLogger(logger_config['logger_name'])
        self.user_inputs = user_input
        self.stop_event = stop_event
        self.rescan_event = rescan_event
        self.exception_q = exception_q
        self.is_running = False
        self.is_main_app_running = is_app_running
        self.run_param_file = run_param_file
        self.last_exception_for_print = ''

        self.need_to_check_printer_ack = False
        self.r2r = None
        self.scanner = None
        self.last_scanned_res = {}
        self.verification_scanner = None
        self.init_hw(logger_config=logger_config)

        # init data:
        self.locations_df = pd.DataFrame()
        self.associator_handler = None
        self.user_inputs['pixels_ids_mapping'] = None if user_input['pixels_mapping_type'] == 'none' else \
            {'type': user_input['pixels_mapping_type'],
             'code_list': pd.read_csv(user_input['pixels_mapping_file'])['external_id'].tolist()}

        try:
            queue_handler = QueueHandler()
            self.do_association_q = queue_handler.get_multiprocess_queue(queue_max_size=ASSOCIATION_Q_SIZE)
            self.associated_q = queue_handler.get_multiprocess_queue(queue_max_size=ASSOCIATION_Q_SIZE)
            self.scan_verification_q = Queue(maxsize=int(self.user_inputs['scan_verification_offset'])+1)

            if self.user_inputs['do_association']:
                initiator_name = os.environ.get('testerStationName')
                kwargs = {'associate_q': self.do_association_q,
                          'associated_status_q': self.associated_q,
                          'stop_event': stop_event,
                          'owner_id': self.user_inputs['owner_id'],
                          'category_id': self.user_inputs['category_id'],
                          'initiator_name': initiator_name,
                          'common_run_name': os.path.basename(logger_config['logger_path'])}
                # check cloud connection:
                AssociatorProcess(owner_id=kwargs['owner_id'])
                self.associator_handler = multiprocessing.Process(target=AssociatorProcess,
                                                                  kwargs=kwargs)

        except Exception as e:
            self.logger.warning(f'exception during TagAssociation init: {e}')
            raise e

    def init_hw(self, logger_config=None):
        # connect to the r2r:
        self.r2r = R2R(logger_config=logger_config, counter_start_idx=self.user_inputs['first_location'], r2r_type=self.user_inputs['r2r_type'], r2r_printer_config=self.user_inputs['r2r_printer_config'])
        # connect to scanner
        self.scanner = Scanner(logger_name=self.logger.name, max_test_time=self.user_inputs['max_test_time'], 
                               n_codes=int(self.user_inputs['asset_num_codes']) + int(self.user_inputs['wiliot_num_codes']))
        if self.user_inputs['do_scan_verification']:
            self.verification_scanner = BarcodeScanner(timeout='500', is_flash=True)

        # connect to printer
        printing_config = Printer.set_printer_config()
        printing_config['passJobName'] = 'BLANK'
        printing_config['failJobName'] = 'line_'
        printing_config["stringBeforeCounter"] = ''
        self.printing = Printer(is_print=self.user_inputs['do_printing'],
                                exception_queue=self.exception_q,
                                logger_name=self.logger.name,
                                printing_config=printing_config)

    def run_app(self):
        try:
            if self.associator_handler is not None:
                self.associator_handler.start()
        except Exception as e:
            self.logger.warning(f'exception during TagAssociation run_app: {e}')
            raise e
        self.run()

    def run(self):
        is_done = False
        self.logger.info('MoveAndScan Start')
        self.is_running = True
        end_of_reel = 0
        max_steps_after_end_of_reel = 2
        while True:
            time.sleep(0)
            cur_time = time.time()
            try:
                if self.stop_event.is_set():
                    self.logger.info('MoveAndScan Stop')
                    self.exit_app()
                    return
                elif self.is_running != self.is_main_app_running():
                    self.is_running = not self.is_running
                    if self.is_running:
                        self.continue_app()
                    else:
                        self.pause_app()
                elif not self.exception_q.empty():
                    time.sleep(0.1)
                    continue

                if self.is_running:
                    # scan
                    time.sleep(0.1)  # wait for r2r to finish moving
                    try:
                        self.scan()
                        self.last_exception_for_print = ''
                    except ValueError as e:
                        if self.user_inputs['rescan_mode'] != 'none':
                            self.logger.warning(f'Do RE-SCAN Got exception during scan: {e}')
                            self.last_exception_for_print = str(e)
                            self.rescan_event.set()
                            while self.rescan_event.is_set():
                                time.sleep(1)
                        else:
                            raise e

                    if self.need_to_check_printer_ack and not self.printing.get_printing_ack(timeout=3):
                        raise Exception('Did not get ack from printer')
                    self.printing.set_printing_type(self.last_scanned_res['scan_status'])
                    # print here

                    # add data to dataframe:
                    self.add_data(new_result=self.last_scanned_res)

                    # add data to scanner verification
                    if self.user_inputs['do_scan_verification']:
                        self.check_scan_verification(self.last_scanned_res)

                    # wait
                    time_to_wait = max([0, float(self.user_inputs['min_test_time']) - (time.time() - cur_time)])
                    time.sleep(time_to_wait)

                    # move
                    if self.user_inputs['is_step_machine'].lower() == 'yes':
                        if end_of_reel >= max_steps_after_end_of_reel:
                            raise R2RException('End of reel reached')
                        if self.r2r.is_r2r_move_during_test():
                            raise R2RException('R2R moved during test!')
                        try:
                            self.r2r.move_r2r()
                        except IndexError as e:
                            is_done = True
                        if self.r2r.is_r2r_move(timeout=float(self.user_inputs['time_to_move'])):
                            end_of_reel = 0
                        else:
                            end_of_reel += 1
                            self.logger.info(f'End of reel reached: continue step number {end_of_reel} out of {max_steps_after_end_of_reel}')

                    if not self.associated_q.empty():
                        self.merge_associated_data()

                    self.need_to_check_printer_ack = True
                    if is_done:
                        time.sleep(30)
                        raise R2RException('End of reel reached')
                else:
                    time.sleep(1)
                

            except Exception as e:
                self.logger.warning(f'MoveAndScan got exception: {e}')
                if not self.exception_q.full():
                    self.exception_q.put(f'MoveAndScan: {e}')
                    self.logger.warning(f'Main: {(traceback.format_exc())}')

    def add_data(self, new_result):
        new_result['wiliot_code'] = new_result['wiliot_code'] if len(new_result['wiliot_code']) > 0 else ['']
        new_result['asset_code'] = new_result['asset_code'] if len(new_result['asset_code']) > 0 else ['']
        added_data = []
        for wiliot_code in new_result['wiliot_code']:
            for asset_code in new_result['asset_code']:
                new_row = {k: v for k, v in new_result.items()}
                new_row['wiliot_code'] = wiliot_code
                new_row['asset_code'] = asset_code
                added_data.append(new_row)

        self.update_last_run_params(added_data[-1])
        self.locations_df = pd.concat([self.locations_df, pd.DataFrame(added_data)], ignore_index=True)

    def update_last_run_params(self, new_data):
        with open(self.run_param_file, 'rb') as f:
            last_run_params = json.load(f)
        last_run_params['last_location'] = new_data['location']
        last_run_params['last_run_name'] = self.user_inputs['run_name']
        if new_data['asset_code']:
            last_run_params['last_asset_id'] = new_data['asset_code']
            last_run_params['last_asset_location'] = new_data['location']
        with open(self.run_param_file, 'w') as f:
            json.dump(last_run_params, f)

    def scan(self):
        self.last_scanned_res = {'location': self.r2r.get_counter(),
                               'wiliot_code': [], 'asset_code': [], 'timestamp': 0,
                               'scan_status': False,
                               'is_associated': False, 'associate_status_code': ''}

        scanned_codes = self.scanner.scan()

        self.check_scanned_codes_and_update_result(scanned_codes)

        # send data to cloud
        if self.user_inputs['do_association'] and self.last_scanned_res['scan_status']:
            if self.do_association_q.full():
                self.logger.warning(f'do_association_q queue is full. discard  {self.last_scanned_res}')
            else:
                self.do_association_q.put(self.last_scanned_res)

        # end of label scanning:
        self.check_end_of_label()

    def get_scan_str(self):
        res_str = f"*{self.last_exception_for_print}*\n"
        res_str += f"wiliot_code {self.last_scanned_res['wiliot_code']}\n"
        res_str += f"asset_code {self.last_scanned_res['asset_code']}\n"
        res_str += f"scan_status {self.last_scanned_res['scan_status']}"
        return res_str

    def disconnect_or_reconnect_scanner(self):
        if self.scanner.is_connected():
            act_str = 'Disconnection'
            self.scanner.disconnect()
            return not self.scanner.is_connected(), act_str
        else: 
            act_str = 'Reconnection'
            self.scanner.reconnect()
            return self.scanner.is_connected(), act_str

    def reprint_previous_label(self):
        self.r2r.r2r.reprint_previous_label()

    def check_scan_verification(self, res):
        # put the asset code in the queue
        if not self.scan_verification_q.full():
            self.scan_verification_q.put(res['asset_code'][0] if res['scan_status'] else None)
        else:
            self.logger.warning(f'scan_verification_q queue is full. discarded {res["asset_code"][0]}')
            raise Exception('scan_verification_q is full')

        # if queue is full, scan code and compare
        if self.scan_verification_q.full():
            exp_code = self.scan_verification_q.get()
            get_code = self.verification_scanner.scan_and_flush()[0]
            self.logger.info(f'Validation Scanner: expected {exp_code}, got {get_code}')
            if exp_code is not None:
                if not exp_code == get_code:
                    if not self.exception_q.full():
                        self.exception_q.put(f'Validation Scanner: expected {exp_code}, but no asset code scanned')
            else:
                if get_code:
                    if not self.exception_q.full():
                        self.exception_q.put(f'Validation Scanner: expected no asset code, got {get_code}')

    def check_scanned_codes_and_update_result(self, codes_in):
        time_now = datetime.datetime.now().timestamp()
        for code in codes_in:
            if is_wiliot_code(code):
                self.last_scanned_res['wiliot_code'].append(code)
                self.last_scanned_res['timestamp'] = time_now
            elif is_asset_code(code, self.user_inputs['asset_code_pattern']):
                self.last_scanned_res['asset_code'].append(code)
                self.last_scanned_res['timestamp'] = time_now
            elif self.user_inputs['scanner_config']['qr_separator'] and self.user_inputs['scanner_config']['qr_separator'] in code:
                code_list = code.split(self.user_inputs['scanner_config']['qr_separator'])
                code_dict = {}
                for key, value in zip(self.user_inputs['scanner_config']['qr_code_mapping'], code_list):
                    code_dict[key] = value
                self.logger.info(f'Parsed QR code: {code_dict}')
                if 'asset_code' not in code_dict:
                    raise ValueError(f'No asset code in scanned QR code: {code} with code dict: {code_dict}')
                asset_code = code_dict.pop('asset_code')
                if is_asset_code(asset_code, self.user_inputs['asset_code_pattern']):
                    for key, value in self.user_inputs['association_config']['association_field_mapping'].items():
                        key_val = code_dict[key]
                        if isinstance(value, list):
                            for v in value:
                                self.last_scanned_res[v] = key_val
                        else:
                            self.last_scanned_res[value] = key_val
                    self.last_scanned_res['asset_code'].append(asset_code)
                    self.last_scanned_res['labels'] = code_dict
                    self.last_scanned_res['timestamp'] = time_now

        if len(self.last_scanned_res['asset_code']) == 0:
            raise ValueError(f'No asset code in scan')
        
        if len(self.last_scanned_res['wiliot_code']) < int(self.user_inputs['wiliot_num_codes']) and self.user_inputs['rescan_mode'] == 'all':
            raise ValueError(f'Not enough Wiliot codes scanned: {self.last_scanned_res["wiliot_code"]}')

        if self.user_inputs['pixels_ids_mapping'] is not None:
            if self.user_inputs['pixels_ids_mapping']['type'] == 'whitelist' and any([code not in self.user_inputs['pixels_ids_mapping']['code_list'] for code in self.last_scanned_res['wiliot_code']]):
                raise ValueError(f'Pixels IDs not in whitelist: {self.last_scanned_res["wiliot_code"]}')
            elif self.user_inputs['pixels_ids_mapping']['type'] == 'blacklist' and any([code in self.user_inputs['pixels_ids_mapping']['code_list'] for code in self.last_scanned_res['wiliot_code']]):
                raise ValueError(f'Pixels IDs in blacklist: {self.last_scanned_res["wiliot_code"]}')
        
        pass_criteria = len(self.last_scanned_res['wiliot_code']) >= int(self.user_inputs['wiliot_num_codes']) \
                and len(self.last_scanned_res['asset_code']) == int(self.user_inputs['asset_num_codes'])
        if self.user_inputs['asset_location'] == 'none':
            self.last_scanned_res['scan_status'] = pass_criteria
        else:
            if (self.user_inputs['asset_location'] == 'first' and is_wiliot_code(codes_in[-1])) or (
                    self.user_inputs['asset_location'] == 'last' and is_wiliot_code(codes_in[0])):
                self.last_scanned_res['scan_status'] = True
                self.logger.info(f'Found all codes for association! '
                                 f'Wiliot:{self.last_scanned_res["wiliot_code"]}, '
                                 f'Asset: {self.last_scanned_res["asset_code"]}')
            else:
                self.last_scanned_res['scan_status'] = False
                raise ValueError(f'Scanned codes from different labels '
                                f'Wiliot:{self.last_scanned_res["wiliot_code"]}, '
                                f'Asset: {self.last_scanned_res["asset_code"]}')

    def check_end_of_label(self):
        if self.last_scanned_res is None:
            return
        if len(self.last_scanned_res['asset_code']) == 0 and len(self.last_scanned_res['wiliot_code']) == 0:
            self.logger.info(f"No codes were read")
        if len(self.last_scanned_res['asset_code']) < ASSET_NUM_CODES:
            self.logger.info(f"Not enough Asset codes were scanned: {self.last_scanned_res['asset_code']}")
        elif len(self.last_scanned_res['asset_code']) > ASSET_NUM_CODES:
            self.logger.info(f"Too many Asset codes were scanned: {self.last_scanned_res['asset_code']}")
        elif len(self.last_scanned_res['wiliot_code']) < WILIOT_MIN_NUM_CODE:
            self.logger.info(f"Not enough Wiliot codes were scanned: {self.last_scanned_res['wiliot_code']}")

    def merge_associated_data(self):
        n = self.associated_q.qsize()
        for _ in range(n):
            associated = self.associated_q.get(block=False)
            print(f'associated: {associated}')
            self.locations_df.loc[(self.locations_df['asset_code'].isin(associated['asset_code'])) & (
                    self.locations_df['wiliot_code'].isin(associated['wiliot_code'])),
                                  ['is_associated', 'associate_status_code']] = [associated['is_associated'],
                                                                                 associated['associate_status_code']]

    def pause_app(self):
        self.need_to_check_printer_ack = False
        self.r2r.pause_app()
        self.scanner.disconnect()

    def continue_app(self):
        self.r2r.continue_app()
        self.scanner.reconnect()

    def exit_app(self):
        try:
            time.sleep(2)
            self.r2r.exit_app()
            self.scanner.disconnect()
            if self.associator_handler is not None:
                self.associator_handler.join(10)
                if self.associator_handler.is_alive():
                    self.logger.warning('associator process is still running')
            if not self.associated_q.empty():
                self.merge_associated_data()
        except Exception as e:
            self.logger.warning(f'MoveAndScan: exit_app: got exception: {e}')
            if not self.exception_q.full():
                self.exception_q.put(f'MoveAndScan: exit_app: {e}')
                self.logger.warning(f'Main: {(traceback.format_exc())}')


    def get_locations_df(self):
        return self.locations_df


if __name__ == '__main__':
    from wiliot_core import set_logger
    RUN_TIME = 60

    tag_assoc_logger_path, tag_assoc_logger = set_logger(app_name='TagAssociation', dir_name='tag_association',
                                                         file_name='association_log')
    stop_event = multiprocessing.Event()
    user_input = {
        'min_test_time': '1.5',
        'max_test_time': '5',
        'time_to_move': '0.5',
        'do_association': 'no',
        'is_step_machine': 'yes',
        'asset_location': 'last',
        'owner_id': 'wiliot-ops',
        'category_id': '86fd9f07-38b0-466f-b717-2e285b803f5c'
    }
    ta = TagAssociation(user_input=user_input, stop_event=stop_event,
                        logger_config={'logger_name': tag_assoc_logger.name, 'logger_path': tag_assoc_logger_path})

    t_i = time.time()
    while time.time() - t_i < RUN_TIME:
        time.sleep(1)
        df = ta.get_locations_df()
        print(f'n unique locations: {len(df)}')
    # stop run
    stop_event.set()

    df = ta.get_locations_df()
    df_path = tag_assoc_logger_path.replace('.log', '_locations_df.csv')
    print(f'saving data at: {df_path}')
    df.to_csv(df_path, index=False)

    print('done')



"""
Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import logging
import time
import threading
import pandas as pd
from enum import Enum

import dash
from dash import dcc
from dash import html
import dash_bootstrap_components as dbc
import dash_daq as daq
from dash.dependencies import Input, Output
from werkzeug.serving import make_server
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from wiliot_core import set_logger


class LevelType(Enum):
    ADVA = 'adv. address'
    EXTERNAL_ID = 'external id'


GREEN_BUTTON_STYLE = {'background-color': 'green',
                      'color': 'white',
                      'height': '50px',
                      'width': '200px'}

RED_BUTTON_STYLE = {'background-color': 'red',
                    'color': 'white',
                    'height': '50px',
                    'width': '200px'}



class YieldPlotting(object):

    def __init__(self, get_data, get_sensors_data, get_app_errors, cmd_q, user_config, logger=None, stop_event=None, tester_name='Yield Tester'):
        """
        :param data_q data queue contains DecryptedTagCollection elements
        :type data_q multiprocessing.Manager().Queue
        """
        if logger is None:
            _, self.logger = set_logger(app_name='YieldPlotting')
        else:
            self.logger = logger
        
        self.tester_name = tester_name.replace('_', ' ').title()
        self.stop_event = stop_event
        self.get_live_df = get_data
        self.cmd_q = cmd_q
        self.user_config = user_config
        self.cur_state = {}
        self.live_df = {}
        self.dash_app = None
        self.server = None
        self.level_type = LevelType.ADVA
        self.get_sensors_data = get_sensors_data
        self.sensors_data = {}
        self.update_sensors()
        self.get_app_errors = get_app_errors
        self.is_run = True

        self.init_live_plot()
        self.logger.info('start application')

    def init_live_plot(self):
        self.logger.info('Starting server connection')
        self.dash_app = dash.Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP, dbc.icons.BOOTSTRAP])
        self.generate_monitor_inlay()
        self.server = self.dash_app.server
        self.init_callbacks()
        self.logger.info('complete init callbacks')

    def init_callbacks(self):

        @self.dash_app.callback(
            Output('placeholder-data_type_switch', 'children'),
            Input("data_type_switch", "value")
        )
        def update_data_type_switch(value):
            self.logger.info('update_data_type_switch')
            try:
                self.level_type = LevelType(value)
                self.logger.info(f'data_type was changed to {self.level_type}')
            except Exception as e:
                self.logger.warning(f'cannot change level type due to {e}')
                raise dash.exceptions.PreventUpdate
            
        @self.dash_app.callback(
            Output('placeholder-window-size', 'children'),
            Input('window-size', 'value')
        )
        def update_plot_mode(value):
            self.user_config['window_size'] = value
            self.logger.info(f'window_size was changed to {value}')
            return None
        
        @self.dash_app.callback(
            Output('placeholder-max-y-axis', 'children'),
            Input('max-y-axis', 'value')
        )
        def update_limit_y_axis(value):
            self.user_config['max_y_axis'] = value
            self.logger.info(f'max_y_axis was changed to {value}')
            return None
        
        @self.dash_app.callback(
            [Output('start-stop-btn', 'children'),
             Output('start-stop-btn', 'color'),],
            Input('start-stop-btn', 'n_clicks')
        )
        def start_stop_run(n_clicks):
            if n_clicks > 0:
                self.is_run = not self.is_run
                if self.cmd_q.full():
                    self.logger.warning(f'start_stop_run: cmd_q is full, discard: {self.cmd_q.get()}')
                self.cmd_q.put('start' if self.is_run else 'pause')
            if self.is_run:
                str_print = 'Run Started'
                style_out = 'danger'
                value = 'Pause Run'
            else: 
                str_print = 'Run Stopped'
                style_out = 'success'
                value = 'Start Run'
            self.logger.info(f'{str_print}')
            return value, style_out

        @self.dash_app.callback(
            [Output('upload-btn', 'children'),
             Output('upload-btn', 'color'),
             Output('upload-btn', 'disabled'),
             Output('start-stop-btn', 'disabled'),
             Output('start-stop-btn', 'style'),
             Output('num-tags-card', 'className'),
             Output('num-adva-card', 'className'),
             Output('num-ex-id-card', 'className')],
            Input('upload-btn', 'n_clicks')
        )
        def upload_and_finish_run(n_clicks):
            style_out = 'secondary'
            value = 'Stop and Upload Data'
            disable_upload = False
            disable_run = False
            run_style = {'height': '40px', 'width': '200px'}
            num_tags_class = 'col bg-light'
            num_adva_class = 'col bg-info text-white'
            num_exid_class = 'col bg-primary text-white'

            if n_clicks > 0:
                self.logger.info('User stop the application')
                style_out = 'light'
                value = 'Test Completed'
                disable_upload = True
                disable_run = True

                run_style['backgroundColor'] = 'lightgrey'
                run_style['borderColor'] = 'lightgrey'
                grey_class = 'col bg-light text-dark'
                num_tags_class = grey_class
                num_adva_class = grey_class
                num_exid_class = grey_class

                if self.cmd_q.full():
                    self.logger.warning(f'upload_and_finish_run: cmd_q is full, discard: {self.cmd_q.get()}')
                self.cmd_q.put('stop')
            return value, style_out, disable_upload, disable_run, run_style, num_tags_class, num_adva_class, num_exid_class

        @self.dash_app.callback(
            [Output('graph', 'figure'),
             Output('num-tags-text', 'children'),
             Output('num-adva-text', 'children'),
             Output('num-ex-id-text', 'children'),
             ] + [Output(f'{sensor_name}-sense', 'children') for sensor_name in self.sensors_data.keys()],
            Input('interval-component', 'n_intervals')
        )
        def update_graphs(_):
            data = self.get_live_df()
            if 0 in data.index:
                data = data.drop(index=0)
            
            if data.empty:
                raise dash.exceptions.PreventUpdate
            sensors_out = self.update_sensors()
            figure_out = self.generate_figures(data)
            n_tags = html.H2(str(data.index[-1] * int(self.user_config.get('matrix_size', 1))), className="card-title")
            n_adva = html.H2(str(data['matrix_advas'].sum()), className="card-title")
            n_ex_id = html.H2(str(data['matrix_external_ids'].sum()), className="card-title")
            return [figure_out, n_tags, n_adva, n_ex_id] + sensors_out

    def update_sensors(self):
        self.sensors_data = self.get_sensors_data()
        if 'temperature' in self.sensors_data.keys() and self.user_config.get('temperature_type', 'C') == 'F':
            self.sensors_data['temperature'] = (self.sensors_data['temperature'] * (9 / 5)) + 32

        sensors_cards = []
        for sensor_type, sensor_data in self.sensors_data.items():
            color = 'black' if float(self.user_config.get(f'min_{sensor_type}', -float('inf'))) <= sensor_data <= float(self.user_config.get(f'max_{sensor_type}', float('inf'))) else 'danger'
            sensors_cards.append(html.P(f'{sensor_data}', className=f"text-{color}"))
        
        return sensors_cards
    
    def generate_figures(self, new_df):
        user_warning = ''    
        matrix_num = new_df.index
        if self.level_type == LevelType.EXTERNAL_ID:
            n_passed = new_df['matrix_external_ids']
        elif self.level_type == LevelType.ADVA:
            n_passed = new_df['matrix_advas']
        else:
            self.logger.warning(f'data type is not supported: {self.level_type}')
            raise dash.exceptions.PreventUpdate

        window_size = int(self.user_config.get('window_size', 1))
        matrix_size = int(self.user_config.get('matrix_size', 1))

        n_tags = matrix_num * int(self.user_config.get('matrix_size', 1))
        yield_all = (n_passed.cumsum() / n_tags) * 100
        yield_current = (n_passed.rolling(window_size).sum() / (window_size * matrix_size)) * 100
        
        # all fig
        fig = make_subplots(rows=1, cols=2, 
                            subplot_titles=[f'Moving Mean Yield - {round(yield_current.iloc[-1], 2)}%', f'Cumulative Yield - {round(yield_all.iloc[-1], 2)}%'],
                            horizontal_spacing=0.1
        )
        max_y_axis = self.user_config.get('max_y_axis')
        fig_info = [
            {'name': 'current', 'x': matrix_num, 'y': yield_current, 'title': 'Number of Matrices'},
            {'name': 'cumulative', 'x': n_tags, 'y': yield_all, 'title': 'Number of Tags'},
        ]

        for ii, sub_data in enumerate(fig_info):
            min_yield = int(self.user_config.get(f'red_line_{sub_data["name"]}', 0))
            if max_y_axis:
                sub_data['y'] = sub_data['y'].clip(upper=float(max_y_axis))

            fig.add_trace(go.Scatter(x=sub_data['x'].to_list(), y=sub_data['y'].to_list(), mode='lines+markers', marker=dict(color='blue')), row=1, col=ii+1)
            fig.add_hline(y=min_yield, line_width=3, line_dash="dash", line_color="red", row=1, col=ii+1)

            fig.update_xaxes(title_text=sub_data['title'], row=1, col=ii+1)
            if sub_data['y'].iloc[-1] < min_yield:
                user_warning += (f' Low {sub_data["name"].capitalize()} Yield!!')
        
        warning_to_print = self.get_app_errors() + user_warning
        fig.update_yaxes(title_text='%')
        
        fig.update_layout(
            title=None,  # remove normal title
            showlegend=False,
            annotations=[
                dict(
                    text=warning_to_print,
                    x=0.5, y=1.05,
                    xref='paper', yref='paper',
                    showarrow=False,
                    font=dict(color='red', size=24),
                    bgcolor='yellow' if warning_to_print else None
                )
            ]
        )
        return fig
    
    def generate_monitor_inlay(self):
        title = html.H2(f"Wiliot {self.tester_name}", className='mt-2', style=dict(display='flex', width='30vw'))
        logo = html.Img(src='https://www.wiliot.com/src/uploads/Wiliotlogo.png', style={"float": "right", "height": 50})
        start_stop_button = html.Div(
            [dbc.Button("Pause Run", id="start-stop-btn",n_clicks=0, color='danger',
                        style={'height': '40px', 'width': '200px'})]
            )
        upload_button = html.Div(
            [dbc.Button("Stop and Upload Data", id="upload-btn",n_clicks=0, color='secondary',
                        style={'height': '40px', 'width': '300px'})]
            )

        radio_button_style = {'padding-left': 10, 'padding-right': 40}

        # Card components
        cards_stat = [
            dbc.Card(id='num-tags-card', className='col bg-light', children=dbc.CardBody(
                [
                    html.Div(id='num-tags-text', className='pe-4'),
                    html.P("Tested Tags", className="card-text"),
                ],
                style=dict(display='flex', width='100%', height='60px')
            )),
            dbc.Card(id='num-adva-card', className='col bg-info text-white', children=dbc.CardBody(
                [
                    html.Div(id='num-adva-text', className='pe-4'),
                    html.P("Adv. Address", className="card-text"),
                ],
                style=dict(display='flex', width='100%', height='60px')
            )),
            dbc.Card(id='num-ex-id-card', className='col bg-primary text-white', children=dbc.CardBody(
                [
                    html.Div(id='num-ex-id-text', className='pe-4'),
                    html.P("External Ids", className="card-text"),
                ],
                style=dict(display='flex', width='100%', height='60px')
            )),
        ]
        sensors_labels = html.P(children=[
            html.Div('nan', id='temperature-sense'), self.user_config.get('temperature_type', 'C'),  html.I(className='bi bi-thermometer-half me-2'), 
            html.P('|',className='px-3'),
            html.Div('nan', id='humidity-sense'), ' %', html.I(className='bi bi-droplet-half me-2'), 
            html.P('|',className='px-3'),
            html.Div('nan', id='light_intensity-sense'), ' Lux', html.I(className='bi bi-lightbulb me-2'),
            ],
            className='col fw-bold', style=dict(display='flex', width='150%')
                                )
        plot_options = [
            html.Div(
                [
                    dcc.RadioItems(id='data_type_switch', 
                                    options=[
                                        {'label': html.Span(LevelType.ADVA.name, style=radio_button_style), 'value': LevelType.ADVA.value},
                                        {'label': html.Span(LevelType.EXTERNAL_ID.name, style=radio_button_style), 'value': LevelType.EXTERNAL_ID.value},
                                        ], 
                                    value=LevelType.EXTERNAL_ID.value, inline=True, className='col', style=dict(display='flex', width='20vw')),
                ],
                className='col fw-bold', style=dict(display='flex', width='100%')
            ),
            html.Div(
                [
                    html.P('Window Size:',className='me-4'),
                    daq.NumericInput(id='window-size', value=int(self.user_config.get('window_size', 1)),
                                     min=0, max=1000),
                ],
                className='col fw-bold', style=dict(display='flex', width='100%')
            ),
            html.Div(
                [
                    html.P('Max Y-Axis:',className='me-4'),
                    daq.NumericInput(id='max-y-axis', value=int(self.user_config.get('max_y_axis', 110)),
                                     min=0, max=500),
                ],
                className='col fw-bold', style=dict(display='flex', width='100%')
            ),
            html.Div(id='placeholder-data_type_switch'),
            html.Div(id='placeholder-window-size'),
            html.Div(id='placeholder-max-y-axis'),
        ]
        plot_style = {'margin': 0,
                      'display': 'flex',
                      'height': '70vh'}
        graph = dcc.Graph(id=f"graph", style=plot_style)

        self.dash_app.layout = dbc.Container(
            [
                dbc.Row([dbc.Col(title)] + [dbc.Col(card) for card in cards_stat] + [dbc.Col(logo)]),
                html.Hr(className='vw-100'),
                dbc.Row([dbc.Col(sensors_labels), dbc.Col(start_stop_button), dbc.Col(upload_button)], className='vw-100'),
                html.Br(),
                dbc.Row([dbc.Col(op) for op in plot_options], className='vw-100 m-0', style={'height': '40px'}),
                dbc.Row(dbc.Col(graph), className='vw-100'),
                dcc.Interval(
                    id='interval-component',
                    interval=1 * 1000,  # in milliseconds
                    n_intervals=0),
            ],
            fluid=True,
        )

    def run(self, host=None, port=None):
        host = self.user_config['host'] if host is None else host
        port = int(self.user_config['port'] if port is None else port)
        self.server = make_server(host, port, self.dash_app.server)
        self.logger.info(f"http://{host}:{port}/")
        logging.getLogger('werkzeug').setLevel(logging.ERROR)
        if self.stop_event is None:
            self.server.serve_forever()
        else:
            all_threads_handles = {}
            server_thread = threading.Thread(target=self.server.serve_forever)
            server_thread.start()
            all_threads_handles['server'] = server_thread
            while not self.stop_event.is_set():
                time.sleep(1)
            self.server.shutdown()
            for handler_name, handler in all_threads_handles.items():
                handler.join(5)
                if handler.is_alive():
                    self.logger.warning(f'failed to stop the {handler_name}')
                else:
                    self.logger.info(f'{handler_name} was stopped')


if __name__ == '__main__':
    from queue import Queue
    my_df = pd.DataFrame()
    x = 0

    def get_data():
        global my_df
        my_df = pd.concat([my_df, pd.DataFrame({'matrix_advas': 1, 'matrix_external_ids': 10, 'trigger_time': 0.0}, index=[0])], ignore_index=True)
        return my_df
    def get_sensors_data():
        global x
        x +=1
        return {'temperature': x, 'humidity': 10*x, 'light_intensity': x*100}
    def get_app_errors():
        return ''
    cmd_q = Queue(maxsize=100)
    stop_event = threading.Event()

    user_config={'port': 8008, 'host': '127.0.0.3',
                 'red_line_current': 50, 'red_line_cumulative': 10,
                 'max_y_axis': '150',
                 "max_temperature": "40",
                 "min_temperature": "10",
                 "temperature_type": "C",
                 "min_humidity": "20",
                 "max_humidity": "90",
                 "min_light_intensity": "0",
                 "max_light_intensity": "1500",

    }
    
    p = YieldPlotting(get_data=get_data, get_sensors_data=get_sensors_data, get_app_errors=get_app_errors, cmd_q=cmd_q, user_config=user_config, stop_event=stop_event, tester_name='conversion_yield_tester')
    p.run(port= 8009, host= "127.0.0.1")



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
from threading import Thread
import logging
import threading
try:
    from tkinter import *
    import pygubu
except Exception as e:
    print(f'could not import tkinter or pygubu: {e}')
import serial.tools.list_ports  # type: ignore
from os.path import abspath
from json import dump
from os.path import join, dirname
from time import sleep
from wiliot_tools.test_equipment.test_equipment import YoctoTemperatureSensor, BarcodeScanner, Attenuator, Tescom, serial_ports
from wiliot_core import WiliotGateway
from wiliot_testers.sample.configs_gui import CONFIGS_DIR

barcodeMutex = threading.Lock()

CONNECT_HW = 'Connect HW'
GO = 'Go'
CONTINUE = 'Continue'
READ = 'Read'
FINISH = 'Finish'

ADD = 'ADD'
REMOVE = 'REMOVE'

GW_CANCEL = '!reset'

# gateway
GW_VERSION = 'Gateway version: '
GW_AVAILABLE_VERSION = 'Available Version: '

# attenuator
ATTENUATION = 'atten'
ATTENUATION_PARAMS = ['attenBle', 'attenLoRa']

BLE = 'Ble'
LORA = 'LoRa'

BARCODES = 'Barcodes'
ATTENUATORS = 'Atten'
CHAMBERS = 'Chambers'

TIME_TO_CLOSE_CHAMBER = 3
MAX_BLE_POWER = 22
MAX_LORA_POWER = 29


class ComConnect(object):
    '''
    classdocs
    '''
    isGui = False
    hwConnected = False
    cur_gw_tbp_version = False
    missing_com_port = False
    wiliotTags = False
    gateway = None
    attenuator = None
    chambers_move_com = ''
    barcodes_move_com = ''
    gwVersion = ''
    reel_id = ''
    gtin = ''
    barcodes_state = ADD
    chambers_state = ADD
    barcodes_serials = {}
    chambers_serials = {}
    atten_serials = {}
    sensors_serials = {}
    used_ports = []
    gw_com_port = []
    barcode_error = []
    chambers_to_close = []
    gw_latest_version = ['']
    gw_update_status = 'disabled'
    temperature_sensor_readings = []
    all_sensors = []
    ttk = None
    logger_dir = ''

    def __init__(self, top_builder=None, new_tag_func=None, update_go=None, default_dict=None, logger=None,
                 logger_dir=None, tk_frame=None):
        '''
        Constructor
        '''
        self.logger_dir = logger_dir
        self.top_builder = top_builder
        self.add_tag_to_test = new_tag_func
        self.update_go_state = update_go
        self.default_dict = default_dict
        self.tk_frame = tk_frame
        if logger is None:
            self.logger = logging.getLogger('sample')
        else:
            self.logger = logger

    def __del__(self):
        if self.gateway is not None and self.is_gw_serial_open():
            self.gateway.exit_gw_api()
            self.gateway = None

        for com_port, barcode in self.barcodes_serials.items():
            if barcode is not None and barcode.is_open():
                barcode.close_port()
        self.barcodes_serials = {}

        for com_port, chamber in self.chambers_serials.items():
            if chamber is not None and chamber.is_connected():
                chamber.open_chamber()
                chamber.close_port()
        self.chambers_serials = {}

        for com_port, atten in self.atten_serials.items():
            if atten != None and 'serial' in atten.keys() and atten['serial'] != None and atten:
                atten['serial'].GetActiveTE().close_port()
        self.atten_serials = {}

    def gui(self):
        self.builder = builder = pygubu.Builder()
        ui_file = join(abspath(dirname(__file__)), 'utils', 'com_connect.ui')
        self.builder.add_from_file(ui_file)

        img_path = join(abspath(dirname(__file__)), '')
        builder.add_resource_path(img_path)
        img_path = join(abspath(dirname(__file__)), 'utils')
        builder.add_resource_path(img_path)

        if self.tk_frame is not None:
            self.ttk = Toplevel(self.tk_frame)
        else:
            self.ttk = Tk()

        self.ttk.title("ComConnect")

        self.mainwindow = self.builder.get_object('mainwindow', self.ttk)

        self.builder.connect_callbacks(self)
        self.isGui = True
        self.find_com_ports_cb()
        self.set_gui_defaults()

        self.ttk.protocol("WM_DELETE_WINDOW", self.close)
        self.ttk.lift()
        self.ttk.attributes("-topmost", True)
        self.ttk.attributes("-topmost", False)

        # self.set_gui_defaults()

        self.ttk.mainloop()

    def set_gui_defaults(self):

        if self.serials_connected(CHAMBERS):
            self.builder.get_object('connect_chambers').configure(text='Disconnect')
            self.builder.get_object('chosenChambers')['state'] = 'disabled'
            self.builder.get_object('availableChambers')['state'] = 'disabled'
            self.builder.get_object('chambers_up')['state'] = 'disabled'
            self.builder.get_object('chambers_down')['state'] = 'disabled'
            self.builder.get_object('addChambers')['state'] = 'disabled'
        else:
            self.builder.get_object('connect_chambers').configure(text='Connect')

        ble_atten = ''
        lora_atten = ''
        for com_port, atten in self.atten_serials.items():
            ble_atten = com_port if atten['type'] == BLE else ble_atten
            lora_atten = com_port if atten['type'] == LORA else lora_atten

        self.builder.get_object('attenComBle').set(ble_atten)
        self.builder.get_object('attenComLoRa').set(lora_atten)
        if self.serials_connected(ATTENUATORS):
            self.builder.get_object('connect_atten').configure(text='Disconnect')
            self.builder.get_object('attenComBle')['state'] = 'disabled'
            self.builder.get_object('attenComLoRa')['state'] = 'disabled'
        else:
            self.builder.get_object('connect_atten').configure(text='Connect')
        pass

    def update_com_gui(self, available_ports, com_ports):
        self.builder.get_object('gwCom')['values'] = available_ports + ['']
        self.builder.get_object('attenComBle')['values'] = available_ports + ['']
        self.builder.get_object('attenComLoRa')['values'] = available_ports + ['']
        self.update_multi_serials(available_ports, BARCODES)
        self.update_multi_serials(available_ports, CHAMBERS)

        self.missing_com_port = False

        self.check_chosen_ports(com_ports)
        self.check_opened_ports()

    def check_chosen_ports(self, com_ports):
        if len(self.gw_com_port) == 0 or self.gw_com_port[0] not in com_ports:
            self.gw_com_port = ['']
            self.builder.get_object('gwCom').set('')
            self.missing_com_port = True

        i = 0
        while i < len(self.atten_serials.keys()):
            port = list(self.atten_serials.keys())[i]
            atten = list(self.atten_serials.values())[i]
            if port != '' and port not in com_ports:
                self.atten_serials.pop(port)
                self.builder.get_object(f"attenCom{atten['type']}").set('')
                self.missing_com_port = True
                continue
            i += 1

        self.check_multi_coms(BARCODES, com_ports)
        self.check_multi_coms(CHAMBERS, com_ports)

    def check_multi_coms(self, obj, com_ports):
        self.builder.get_object(f'chosen{obj}').delete(0, END)
        ports = getattr(self, f'{obj.lower()}_serials')
        for port in ports.keys():
            if port in com_ports:
                self.builder.get_object(f'chosen{obj}').insert(END, port)
            else:
                self.missing_com_port = True

    def check_opened_ports(self):
        if self.is_gui_opened():
            self.check_gw_open()
            self.check_multi_open(BARCODES)
            self.check_multi_open(CHAMBERS)
            self.check_atten_open()

    def check_multi_open(self, obj):
        if self.serials_connected(obj):
            self.builder.get_object(f'connect_{obj.lower()}').configure(text='Disconnect')
            self.builder.get_object(f'chosen{obj}')['state'] = 'disabled'
            self.builder.get_object(f'available{obj}')['state'] = 'disabled'
            self.builder.get_object(f'{obj.lower()}_up')['state'] = 'disabled'
            self.builder.get_object(f'{obj.lower()}_down')['state'] = 'disabled'
            self.builder.get_object(f'add{obj}')['state'] = 'disabled'
        else:
            self.builder.get_object(f'connect_{obj.lower()}').configure(text='Connect')
            self.builder.get_object(f'chosen{obj}')['state'] = 'normal'
            self.builder.get_object(f'available{obj}')['state'] = 'normal'
            self.builder.get_object(f'{obj.lower()}_up')['state'] = 'normal'
            self.builder.get_object(f'{obj.lower()}_down')['state'] = 'normal'
            self.builder.get_object(f'add{obj}')['state'] = 'normal'

    def check_gw_open(self):
        if len(self.gw_com_port) > 0 and self.gw_com_port[0] != '':
            self.builder.get_object('gwCom').set(self.gw_com_port[0])
        if self.is_gw_serial_open():
            self.builder.get_object('connect_gw').configure(text='Disconnect')
            self.builder.get_object('gwCom')['state'] = 'disabled'
            self.builder.get_object('version').configure(text=GW_VERSION + self.gwVersion[0])
            self.builder.get_object('latestVersion').configure(text=GW_AVAILABLE_VERSION + self.gw_latest_version[0])
            self.builder.get_object('update_gw')['state'] = self.gw_update_status
        else:
            self.builder.get_object('connect_gw').configure(text='Connect')
            self.builder.get_object('gwCom')['state'] = 'normal'
            self.builder.get_object('version').configure(text=GW_VERSION)
            self.builder.get_object('latestVersion').configure(text=GW_AVAILABLE_VERSION)
            self.builder.get_object('update_gw')['state'] = 'disabled'

    def check_atten_open(self):
        connected = False
        for com, atten in self.atten_serials.items():
            if atten['serial'] is not None and atten['serial'].GetActiveTE().is_open():
                connected = True
        if connected:
            self.builder.get_object(f'connect_atten').configure(text='Disconnect')
            self.builder.get_object(f'attenComLoRa')['state'] = 'disabled'
            self.builder.get_object(f'attenComBle')['state'] = 'disabled'
        else:
            self.builder.get_object(f'connect_atten').configure(text='Connect')
            self.builder.get_object(f'attenComLoRa')['state'] = 'normal'
            self.builder.get_object(f'attenComBle')['state'] = 'normal'

    def choose_com_ports(self):
        default_dict = self.default_dict
        com_ports = [comport.device for comport in serial.tools.list_ports.comports()]
        if len(com_ports) == 0:
            com_ports = [s.name for s in serial.tools.list_ports.comports()]
        if 'gw' in default_dict.keys() and default_dict['gw'] in com_ports:
            self.gw_com_port = [default_dict['gw']]
            self.used_ports.append(default_dict['gw'])
        else:
            self.gw_com_port = ['']
            self.missing_com_port = True
        if 'atten' in default_dict.keys() and BLE in default_dict['atten'].keys() and default_dict['atten'][BLE] \
                in com_ports:
            self.atten_serials[default_dict['atten'][BLE]] = {}
            self.atten_serials[default_dict['atten'][BLE]]['type'] = BLE
            self.atten_serials[default_dict['atten'][BLE]]['serial'] = None
            self.used_ports.append(default_dict['atten'][BLE])
        elif 'atten' in default_dict.keys() and BLE in default_dict['atten'].keys() \
                and default_dict['atten'][BLE].strip() != '':
            self.missing_com_port = True
        if 'atten' in default_dict.keys() and LORA in default_dict['atten'].keys() and default_dict['atten'][LORA] \
                in com_ports:
            self.atten_serials[default_dict['atten'][LORA]] = {}
            self.atten_serials[default_dict['atten'][LORA]]['type'] = LORA
            self.atten_serials[default_dict['atten'][LORA]]['serial'] = None
            self.used_ports.append(default_dict['atten'][LORA])
        elif 'atten' in default_dict.keys() and LORA in default_dict['atten'].keys() \
                and default_dict['atten'][LORA].strip() != '':
            self.missing_com_port = True

        if 'barcodes' in default_dict.keys():
            self.barcodes_serials = dict.fromkeys([barcode for barcode in default_dict['barcodes'] if barcode
                                                   in com_ports], None)
            self.used_ports += list(self.barcodes_serials.keys())

        if 'chambers' in default_dict.keys():
            self.chambers_serials = dict.fromkeys([chamber for chamber in default_dict['chambers'] if chamber
                                                   in com_ports], None)
            self.used_ports += list(self.chambers_serials.keys())

        if 'temperature_sensors' in default_dict.keys():
            self.sensors_serials = dict.fromkeys([sensor for sensor in default_dict['temperature_sensors']], None)

        missing_barcodes = []
        missing_chambers = []
        if 'barcodes' in default_dict.keys():
            missing_barcodes = [barcode for barcode in default_dict['barcodes'] if barcode not in com_ports]
        if 'chambers' in default_dict.keys():
            missing_chambers = [chamber for chamber in default_dict['chambers'] if chamber not in com_ports]
        if any(missing_barcodes + missing_chambers):
            self.missing_com_port = True

        return self.missing_com_port

    def init_gw(self):
        self.logger.info('Initializing Gateway object')
        return WiliotGateway(logger_name=self.logger.name, mp_reset_time_upon_gw_start=True,
                             np_max_packet_in_buffer_before_error=10,
                             log_dir_for_multi_processes=join(dirname(self.logger_dir), 'gw_listener_logs'))

    def connect_all(self, gui=True):
        if self.gateway is None:
            self.gateway = self.init_gw()
        if not self.is_gw_serial_open():
            success = self.connect_gw(gui)
            if not success:
                return
        if not self.serials_connected(ATTENUATORS):
            self.connect_atten(gui)
        if not self.serials_connected(BARCODES):
            self.connect_barcodes(gui)
        if not self.serials_connected(CHAMBERS):
            self.connect_chambers(gui)
        self.connect_temperature_sensor()
        self.hwConnected = True

    def connect_gw(self, gui=True, disconnect=False):
        if self.gateway is None:
            self.gateway = self.init_gw()
        if not self.is_gw_serial_open() and not disconnect:
            if len(self.gw_com_port) == 0 or self.gw_com_port[0].strip() == '':
                self.popup_message('No default com port for GW, please choose GW com port.', title='Error', log='error')
                return False
            com_port = self.gw_com_port[0]
            self.gateway.open_port(port=com_port, baud=921600)
            if self.is_gw_serial_open():
                self.start_listener(not_print_str=True)
                self.logger.info(f'GW is connected on port: {com_port}.')
                self.gateway.reset_gw()
                sleep(1)
                version = self.gateway.get_gw_version()
                self.gwVersion = version
                self.gw_latest_version = latest_version = self.gateway.get_latest_version_number()
                cur_version = int(version[0].replace('.', ''))
                self.gw_update_status = 'normal' if cur_version < int(latest_version[0].replace('.', '')) \
                    else 'disabled'
            else:
                self.logger.error(f'Error connecting to GW on port: {com_port}.')
                return False
        else:
            self.gateway.close_port()
            self.builder.get_object('connect_gw').configure(text='Connect')
            self.builder.get_object('version').configure(text=GW_VERSION)
            self.builder.get_object('latestVersion').configure(text=GW_AVAILABLE_VERSION)
            self.builder.get_object('gwCom')['state'] = 'normal'
        if gui:
            self.check_gw_open()

        return True

    def connect_barcodes(self, gui=True):
        self.connect_multi_serials(BARCODES, gui=gui)

    def connect_chambers(self, gui=True):
        self.connect_multi_serials(CHAMBERS, gui=gui)

    def connect_atten(self, gui=True):
        is_connected = self.connect_multi_serials(ATTENUATORS, gui=gui)
        if gui:
            atten_state = 'disabled' if is_connected else 'normal'
            self.builder.get_object('attenComLoRa')['state'] = atten_state
            self.builder.get_object('attenComBle')['state'] = atten_state

    def connect_temperature_sensor(self):
        self.all_sensors = []
        self.temperature_sensor_readings = []
        for sensor_name in self.sensors_serials.keys():
            self.temperature_sensor_readings.append([])
            try:
                sensor_temp = YoctoTemperatureSensor()
                is_temp_sensor_connected = sensor_temp.connect(target=sensor_name)
                if is_temp_sensor_connected:
                    self.logger.info('Temperature Sensor {} is connected'.format(sensor_temp.get_sensor_name()))
                    self.all_sensors.append(sensor_temp)
                else:
                    self.popup_message('Could not connect to Temperature Sensor according to '
                                       'the following name: {}'.format(sensor_name), title='Error', log='error')
                    self.all_sensors.append(None)
                    raise ConnectionError("Could not establish Temperature sensor connection")
            except Exception as e:
                self.logger.info('while connecting to the Temperature Sensor the following error occurs : {}'.format(e))
                self.all_sensors.append(None)
                raise Exception(f'Could not establish Temperature sensor connection due to {e}')

    def read_temperature_sensor(self):
        for i, sensor in enumerate(self.all_sensors):
            if sensor is not None:
                self.temperature_sensor_readings[i].append(sensor.get_temperature())

    def connect_multi_serials(self, obj, gui=True):
        serials = getattr(self, f'{obj.lower()}_serials')
        is_connected = False
        if self.serials_connected(obj):
            self.close_serials(obj, serials)
            self.builder.get_object(f'connect_{obj.lower()}').configure(text='Connect')
            try:
                self.builder.get_object(f'chosen{obj}')['state'] = 'normal'
                self.builder.get_object(f'available{obj}')['state'] = 'normal'
                self.builder.get_object(f'add{obj}')['state'] = 'normal'
            except:
                pass
        elif len(serials.keys()) > 0:
            self.open_serials(obj, serials)
            if self.serials_connected(obj) and gui:
                # if gui:
                is_connected = True
                self.builder.get_object(f'connect_{obj.lower()}').configure(text='Disconnect')
                try:
                    self.builder.get_object(f'chosen{obj}')['state'] = 'disabled'
                    self.builder.get_object(f'available{obj}')['state'] = 'disabled'
                    self.builder.get_object(f'add{obj}')['state'] = 'disabled'
                except:
                    pass
        if gui:
            self.find_com_ports_cb()
        # self.update_go_state()
        return is_connected

    def open_serials(self, obj, serials):
        threads = []
        for com_port, com_serial in serials.items():
            if 'barcode' in obj.lower():
                if com_serial is not None and com_serial.is_open():
                    continue
                com_serial = BarcodeScanner(com_port=com_port, log_type='LOG_NL')
                if com_serial.is_open():
                    # self.used_ports.append(com_port)
                    serials[com_port] = com_serial
            elif 'chamber' in obj.lower():
                if com_serial is not None and com_serial.is_connected():
                    continue
                temp_thread = Thread(target=self.connect_chamber, args=([com_port, serials]))
                temp_thread.start()
                threads.append(temp_thread)
            elif 'atten' in obj.lower():
                if com_serial['serial'] is not None or com_port.strip() == '':
                    # 8 if serial['serial']!=None and
                    # serial['serial'].GetActiveTE().s.is_open():
                    continue
                com_serial = Attenuator('API', comport=com_port)
                if com_serial.GetActiveTE().is_open():
                    # self.used_ports.append(com_port)
                    serials[com_port]['serial'] = com_serial

        for thread in threads:
            thread.join()

    def close(self):
        if self.is_gw_serial_open() and self.serials_connected(ATTENUATORS) and self.serials_connected(BARCODES):
            self.hwConnected = True
            self.enable_hw_connected()
        if self.isGui:
            self.isGui = False
            self.ttk.destroy()
        return self.hwConnected

    def save(self):
        self.default_dict['gw'] = self.gw_com_port[0]
        self.default_dict['atten'] = {}
        for com_port, atten in self.atten_serials.items():
            self.default_dict['atten'][atten['type']] = com_port
        self.default_dict['barcodes'] = list(self.barcodes_serials.keys())
        self.default_dict['chambers'] = list(self.chambers_serials.keys())
        with open(join(CONFIGS_DIR, '.defaults.json'), 'w+') as defaultComs:
            dump(dict(self.default_dict), defaultComs, indent=4)

        self.logger.info(f'Com ports saved successfully.')

    def focus_available(self, obj):
        self.builder.get_object(f'add{obj}').configure(text='>')
        setattr(self, f'{obj.lower()}_state', ADD)

    def focus_chosen(self, obj):
        self.builder.get_object(f'add{obj}').configure(text='<')
        setattr(self, f'{obj.lower()}_state', REMOVE)
        setattr(self, f'{obj.lower()}_move_com', '')

    def add_barcode(self):
        if getattr(self, f'{BARCODES.lower()}_state') == ADD:
            com_chosen = self.builder.get_object(f'available{BARCODES}').get(ACTIVE)
            try:
                temp_barcode = BarcodeScanner(com_port=com_chosen)
            except Exception as e:
                self.popup_message(f'Could NOT connect. {e}', title='Error', log='error')
        self.add(BARCODES)
        self.find_com_ports_cb()

    def add_chamber(self):
        self.add(CHAMBERS)
        self.find_com_ports_cb()

    def add(self, obj):
        if getattr(self, f'{obj.lower()}_state') == ADD:
            sending = self.builder.get_object(f'available{obj}')
            receiving = self.builder.get_object(f'chosen{obj}')
        else:
            sending = self.builder.get_object(f'chosen{obj}')
            receiving = self.builder.get_object(f'available{obj}')

        com_list = list(sending.get(0, END))
        com_chosen = sending.get(ACTIVE)
        receiving.insert(END, com_chosen)
        com_index = com_list.index(com_chosen)
        sending.delete(com_index, com_index)

        serials = getattr(self, f'{obj.lower()}_serials')
        com_ports = self.builder.get_object(f'chosen{obj}').get(0, END)
        old_ports = [port for port in serials.keys() if port not in com_ports]
        new_serials = dict(zip(com_ports, [None] * len(com_ports)))
        self.used_ports = [port for port in self.used_ports if port not in old_ports] + list(com_ports)
        setattr(self, f'{obj.lower()}_serials', new_serials)

    def chamber_up(self):
        self.up(CHAMBERS)

    def chamber_down(self):
        self.down(CHAMBERS)

    def barcode_up(self):
        self.up(BARCODES)

    def barcode_down(self):
        self.down(BARCODES)

    def up(self, obj):
        com_list = list(self.builder.get_object(f'chosen{obj}').get(0, END))
        if getattr(self, f'{obj.lower()}_move_com') == '':
            chosen_com = self.builder.get_object(f'chosen{obj}').get(ACTIVE)
            setattr(self, f'{obj.lower()}_move_com', chosen_com)
        else:
            chosen_com = getattr(self, f'{obj.lower()}_move_com')
        if chosen_com != '':
            com_index = com_list.index(chosen_com)
            if com_index > 0:
                com_list.pop(com_list.index(chosen_com))
                com_list.insert(com_index - 1, chosen_com)
                self.builder.get_object(f'chosen{obj}').delete(0, END)
                for com in com_list:
                    self.builder.get_object(f'chosen{obj}').insert(END, com)
                self.builder.get_object(f'chosen{obj}').select_set(com_index - 1)

    def down(self, obj):
        com_list = list(self.builder.get_object(f'chosen{obj}').get(0, END))
        if getattr(self, f'{obj.lower()}_move_com') == '':
            chosen_com = self.builder.get_object(f'chosen{obj}').get(ACTIVE)
            setattr(self, f'{obj.lower()}_move_com', chosen_com)
        else:
            chosen_com = getattr(self, f'{obj.lower()}_move_com')
        if chosen_com != '':
            com_index = com_list.index(chosen_com)
            if com_index < (len(com_list) - 1):
                com_list.pop(com_list.index(chosen_com))
                com_list.insert(com_index + 1, chosen_com)
                self.builder.get_object(f'chosen{obj}').delete(0, END)
                for com in com_list:
                    self.builder.get_object(f'chosen{obj}').insert(END, com)
                self.builder.get_object(f'chosen{obj}').select_set(com_index + 1)

    def update_multi_serials(self, available_ports, obj):
        self.builder.get_object(f'available{obj}').delete(0, END)
        for port in available_ports:
            if port not in self.used_ports:
                self.builder.get_object(f'available{obj}').insert(END, port)

    def update_atten_serials(self, available_ports):
        for com, item in self.atten_serials.items():
            if item['serial'] is None or not item['serial'].is_open():
                self.builder.get_object(f'attenCom{item["type"]}')['values'] = available_ports
            else:
                self.builder.get_object(f'attenCom{item["type"]}').set(com)

    def connect_chamber(self, com_port, serials):
        com_serial = Tescom(com_port)
        if com_serial.is_connected():
            # self.used_ports.append(com_port)
            serials[com_port] = com_serial
            if not com_serial.is_door_open():
                com_serial.open_chamber()

    def close_serials(self, obj, serials):
        if 'atten' in obj.lower():
            for serial in serials.values():
                # serial['serial'].GetActiveTE.s.close_port()
                if serial['serial'] is not None and serial['serial'].GetActiveTE().is_open():
                    serial['serial'].GetActiveTE().close_port()
                serial['serial'] = None
                # self.used_ports.remove(com_port)
        else:
            for com_serial in serials.values():
                if 'chamber' in obj.lower():
                    com_serial.open_chamber()
                com_serial.close_port()
                # self.used_ports.remove(com_port)
                # serials.pop(com_port)

    def serials_connected(self, obj):
        serials = getattr(self, f'{obj.lower()}_serials')
        if 'atten' in obj.lower():
            serials = dict(zip(serials.keys(), [atten['serial'] for atten in serials.values()]))
        connected_serials = 0
        for com_port, com_serial in serials.items():
            if com_serial is not None:
                if 'barcode' in obj.lower() and com_serial.is_open():
                    connected_serials += 1
                if 'chamber' in obj.lower() and com_serial.is_connected():
                    connected_serials += 1
                if 'atten' in obj.lower():
                    connected_serials += 1
            # if com_port.strip()=='':
            #     connected_serials += 1
        if connected_serials > 0 and connected_serials == len(serials.keys()):
            return True
        else:
            return False

    def read_barcode(self, scanner_index=0, close_chamber=False, add_to_test=False, n_try=5):
        full_data, reel_id, gtin = None, None, None
        if len(list(self.barcodes_serials.values())) == 0:
            self.logger.error('Trying to read barcode but no scanner were connected')
            return None, None
        scanner = list(self.barcodes_serials.values())[scanner_index]
        for i in range(n_try):
            full_data, cur_id, reel_id, gtin = scanner.scan_ext_id()
            if full_data is None and cur_id is None and reel_id is None and full_data is None:
                continue
            break

        reel_id = reel_id if reel_id is not None else full_data
        gtin = gtin if gtin is not None else ''
        if reel_id is not None:
            self.reel_id = reel_id
            self.gtin = gtin
        if full_data is None:
            barcodeMutex.acquire()
            if close_chamber:
                self.barcode_error.append(scanner_index)
            barcodeMutex.release()
            return None, None

        if not close_chamber:
            reel_id_obj = self.top_builder.tkvariables.get('reelId')
            reel_id_obj.set(self.reel_id)

        else:
            success = self.add_tag_to_test(full_data, reel_id, scanner_index=scanner_index, add_to_test=add_to_test)
            if success:
                self.chambers_to_close.append(scanner_index)

        return full_data, gtin + reel_id

    def get_all_scanners_index(self):
        return list(range(len(self.barcodes_serials.values())))

    def read_scanners_barcodes(self, indexes=()):
        if len(indexes) == 0:
            indexes = self.get_all_scanners_index()
        scanner_threads = []
        self.barcode_error = []
        self.chambers_to_close = []

        for i in indexes:
            t = threading.Thread(target=self.read_barcode, args=(i, True, True))
            scanner_threads.append(t)
            t.start()
        for i in range(len(scanner_threads)):
            t = scanner_threads[i]
            t.join()

        read_message = ''
        title = 'Warning'
        font = 18
        if len(self.barcode_error) > 0:
            read_message += f'Error reading external ID from chambers {self.barcode_error},' \
                            f' try repositioning the tags.\n'
            title = 'Error'
            font = 16

        if len(self.chambers_to_close) > 0:
            read_message += f'Chambers are closing!!\nWatch your hands!!!'

        self.popup_message(read_message, title, ("Helvetica", font), title.lower())

        if len(self.chambers_to_close) > 0:
            self.close_chambers(self.chambers_to_close)
        sleep(TIME_TO_CLOSE_CHAMBER)
        self.update_go_state()

    def enable_hw_connected(self):
        self.top_builder.get_object('read_qr')['state'] = 'normal'
        self.top_builder.get_object('reelId')['state'] = 'normal'
        if self.top_builder.tkvariables.get('go').get() == CONNECT_HW:
            self.top_builder.tkvariables.get('go').set(READ)
            self.top_builder.get_object('go')['state'] = 'disabled'

    def set_attenuation(self, params):
        if params.get('run_configs', {}).get('isTestSuite',{}):
            return
        for com_port, atten in self.atten_serials.items():  # BLE and/or Lora
            try:
                value = params[atten['type'].lower() + 'Attenuation']
                attenuation = atten['serial'].GetActiveTE().Setattn(float(value))
                self.logger.info(f"{atten['type']} Attenuation set to: {str(attenuation).strip()}")
            except Exception as e:
                err = f"{atten['type']} Attenuator error: try reconnect the attenuator [{e}].\n" \
                      f"or params are not correct: {params.keys()}: [{e}]"
                self.logger.error(err)
                raise Exception(err)

    def connect_and_close(self):
        self.connect_all()
        self.close()

    def get_reel_id(self):
        return self.reel_id

    def get_gtin(self):
        return self.gtin

    def get_reel_external(self):
        return self.gtin + self.reel_id

    def is_gw_serial_open(self):
        if self.gateway is None:
            return False
        serial_open, _, _ = self.gateway.get_connection_status()
        return serial_open

    def is_gui_opened(self):
        return self.isGui

    def get_gw_version(self):
        return self.gwVersion[0]

    def update_gw(self):  #CB
        self.gateway.update_version()

    def cancel_gw_commands(self):
        # self.gateway.write(GW_CANCEL)
        self.gateway.reset_gw()
        # self.gateway.stop_continuous_listener()
        # self.gateway.reset_buffer()
        sleep(0.1)

    def start_listener(self, not_print_str=True):
        # self.gateway.start_continuous_listener(not_print_str)
        self.gateway.start_continuous_listener()

    def is_hw_connected(self):
        return self.hwConnected

    def get_chambers(self):
        return list(self.chambers_serials.values())

    def open_chambers(self, indexes=()):
        chambers_threads = []
        chambers = list(self.chambers_serials.values())
        if len(indexes) == 0:
            indexes = list(range(len(chambers)))
        for index in indexes:
            if len(chambers) > index and chambers[index] is not None:
                temp_thread = Thread(target=chambers[index].open_chamber, args=())
                temp_thread.start()
                chambers_threads.append(temp_thread)

        for thread in chambers_threads:
            thread.join()

    def close_chambers(self, indexes=()):
        chambers = list(self.chambers_serials.values())
        chambers_threads = []
        if len(indexes) == 0:
            indexes = list(range(len(chambers)))
        for index in indexes:
            if len(chambers) > index and chambers[index] is not None:
                t = threading.Thread(target=chambers[index].close_chamber, args=())
                chambers_threads.append(t)
                t.start()
        for thread in chambers_threads:
            thread.join()

    def get_num_of_barcode_scanners(self):
        return len(self.barcodes_serials.keys())

    def get_error_barcode(self):
        return self.barcode_error

    def get_default_dict(self):
        return self.default_dict

    def popup_message(self, msg, title='Error', font=("Helvetica", 10), log='info', bg=None, tk_frame=None):
        if tk_frame:
            popup = Toplevel(tk_frame)
        else:
            popup = Tk()
        popup.eval('tk::PlaceWindow . center')
        popup.wm_title(title)
        if bg is not None:
            popup.configure(bg=bg)
        getattr(self.logger, log)(f'{title} - {msg}')

        def popup_exit():
            popup.destroy()

        label = Label(popup, text=msg, font=font)
        label.pack(side="top", fill="x", padx=10, pady=10)
        b1 = Button(popup, text="Okay", command=popup_exit)
        b1.pack(padx=10, pady=10)
        popup.mainloop()

    # ############## GUI Callbacks  #######################

    def find_com_ports_cb(self, *args):
        com_ports = serial_ports()
        available_ports = [com_port for com_port in com_ports if com_port not in self.used_ports]

        self.update_com_gui(available_ports, com_ports)

    def focus_barcode_available_cb(self, *args):
        self.focus_available(BARCODES)

    def focus_barcode_chosen_cb(self, *args):
        self.focus_chosen(BARCODES)

    def focus_chamber_available_cb(self, *args):
        self.focus_available(CHAMBERS)

    def focus_chamber_chosen_cb(self, *args):
        self.focus_chosen(CHAMBERS)

    def choose_gw_cb(self, *args):
        if len(self.gw_com_port) > 0 and self.gw_com_port[0] != '':
            self.used_ports.pop(self.used_ports.index(self.gw_com_port[0]))
        self.gw_com_port = [self.builder.get_object('gwCom').get()]
        self.used_ports.append(self.gw_com_port[0])

    def choose_ble_atten_cb(self, *args):
        ble_com = self.builder.get_object('attenComBle').get()
        ble_last_com = [com for com, item in self.atten_serials.items() if item['type'] == BLE]
        if len(ble_last_com) > 0:
            self.atten_serials.pop(ble_last_com[0])
            self.used_ports.pop(self.used_ports.index(ble_last_com[0]))
        if ble_com.strip() != '':
            self.atten_serials[ble_com] = {}
            self.atten_serials[ble_com]['type'] = BLE
            self.atten_serials[ble_com]['serial'] = None
            self.used_ports.append(ble_com)

    def choose_lora_atten_cb(self, *args):
        lora_com = self.builder.get_object('attenComLoRa').get()
        lora_last_com = [com for com, item in self.atten_serials.items() if item['type'] == LORA]
        if len(lora_last_com) > 0:
            self.atten_serials.pop(lora_last_com[0])
            self.used_ports.pop(self.used_ports.index(lora_last_com[0]))
        if lora_com.strip() != '':
            self.atten_serials[lora_com] = {}
            self.atten_serials[lora_com]['type'] = LORA
            self.atten_serials[lora_com]['serial'] = None
            self.used_ports.append(lora_com)



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import json
import os
import pathlib
import time
from datetime import datetime
try:
    from tkinter import messagebox
except Exception as e:
    print(f'could not import tkinter: {e}')
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from wiliot_core import WiliotGateway, set_logger, PacketList, DataType, ActionType
from wiliot_tools.utils.wiliot_gui.wiliot_gui import *

from wiliot_tools.test_equipment.test_equipment import EquipmentError, Attenuator
from wiliot_testers.wiliot_tester_log import WiliotTesterLog

# global vars go here :)
global LORA_ATTN_COM
global BLE_ATTN_COM
global STABILIZATION_RANGE_MODIFIER  # Max differance between points that fit the criteria
global MIN_CONSECUTIVE_POINTS_MODIFIER  # How many points in a row should be in order to recommend a value
global OPTIMAL_ATTENUATION_MODIFIER  # By how much attenuation to increase after finding stabilisation point

LORA_ATTN_COM = "COM3"
BLE_ATTN_COM = "COM6"
STABILIZATION_RANGE_MODIFIER = 1
MIN_CONSECUTIVE_POINTS_MODIFIER = 5
OPTIMAL_ATTENUATION_MODIFIER = 2

run_default_values = {
    'time_profile_on': '5',
    'time_profile_period': '15',
    'lora_value': '20',
    'iter_length': '90',
    'energy_pattern_value': '50',
    'ble_sweep_start': 0,
    'ble_sweep_end': 30,
}

multiprocessing = True
now = datetime.now()
run_start_time = now.strftime("%m-%d-%Y_%H-%M-%S")


class BeaconCalibration:

    def __init__(self):
        # Initialises logger and GUI
        self.init_name_path()
        self.init_gui()
        self.gw_logger = WiliotTesterLog(run_name="beacon_calibration_testing")
        self.gw_logger.set_logger(self.gw_log_path, tester_name="Wiliot tester")

        # Initialises the hardware
        self.attn_enable = True
        self.gw_obj = None
        self.attn_obj_lora = None
        self.attn_obj_ble = None
        self.init_hw()
        self.all_results = pd.DataFrame({'Adva': [], 'Frequency': [], 'TBP': [], 'Atten': []})
        self.gw_config()
        self.run_sweep()
        self.process_result()
        self.analysis_plot()

    def init_name_path(self):
        self.results_path = os.path.join('', 'Beacon_calibration_runs',
                                         ('Beacon_calibration_run_{}'.format(run_start_time)))
        if not os.path.exists(self.results_path):
            pathlib.Path(self.results_path).mkdir(parents=True, exist_ok=True)

        self.gw_log_path = os.path.join('', 'Beacon_calibration_logs')  # Gateway log location

        logger_path, self.logger = set_logger(app_name='BeaconCalibration', dir_name='Beacon_calibration_logs',
                                              file_name='Beacon_calibration_test_log')
        self.logger.critical("I love you, unassuming user")

    def init_hw(self):

        self.gw_obj = WiliotGateway(auto_connect=True, logger_name=self.gw_logger.gw_logger.name, verbose=False)
        try:
            time.sleep(1)
            if self.gw_obj.connected:
                self.gw_obj.reset_gw()
                self.gw_obj.reset_buffer()
                self.gw_obj.is_gw_alive()
                self.gw_obj.start_continuous_listener()

                self.gw_logger.gw_logger.info('GW config finished')
            else:
                raise EquipmentError('Gateway Error - Verify WiliotGateway connection')
        except Exception as e:
            self.gw_logger.gw_logger.warning(e)
            self.gw_obj.close_port()
            raise EquipmentError('Gateway Error - Verify WiliotGateway connection')

        if self.attn_enable:
            self.attn_obj_lora = Attenuator(ATTN_type='API', comport=LORA_ATTN_COM).GetActiveTE()
            attn = self.lora_value
            value = self.attn_obj_lora.Setattn(attn=attn)
            if value != attn:
                raise Exception(f'Error setting attenuation: new : {attn} current: {value}')

    def gw_config(self):
        self.gw_obj.config_gw(energy_pattern_val=self.energy_pattern_val,
                              time_profile_val=[self.time_profile_on_val, self.time_profile_period_val])

    def run_sweep(self):
        self.attn_obj_ble = Attenuator(ATTN_type="API", comport=BLE_ATTN_COM).GetActiveTE()
        self.gw_obj.start_continuous_listener()
        sweep_attns = range(self.ble_sweep_start, self.ble_sweep_end, 1)
        self.attenuation_list = []
        self.all_results = PacketList()
        for attn in sweep_attns:
            # set atten
            self.attn_val = self.attn_obj_ble.Getattn()
            self.attn_val = int(self.attn_obj_ble.Setattn(attn=attn))
            self.attenuation_list.append(attn)
            self.logger.info(f'Current setting attenuation: new : {attn} current: {self.attn_val}')

            self.gw_obj.reset_listener()
            self.gw_obj.config_gw(start_gw_app=True)
            self.daq()
            is_stopped = self.gw_obj.stop_gw_app()
            if not is_stopped:
                raise Exception('could not send stop to GW')
        return self.attenuation_list, self.attn_val

    def daq(self):
        time.sleep(2)  # possibly useless, check
        start_time = time.time()  # Records starting time
        max_duration = self.iter_duration
        while True:
            time.sleep(0)

            if self.gw_obj.is_data_available():
                cur_packet_list = self.gw_obj.get_packets(action_type=ActionType.ALL_SAMPLE,
                                                          data_type=DataType.PACKET_LIST)

                if len(cur_packet_list) > 0:
                    for p in cur_packet_list:
                        p.add_custom_data({'attenuation': self.attn_val})
                        self.logger.info("Would you look at that! , i found this packet: {}"
                                         .format(p.packet_data['raw_packet']))
                        # self.logger.info('got packet {}'.format(p.packet_data['raw_packet']))
                        self.all_results.append(p)
            if time.time() - start_time > max_duration:
                break
        self.logger.info('Iteration ended!')

    def process_result(self):
        self.tbp_df = pd.DataFrame()  # creates an empty dataframe

        df = self.all_results.get_df(add_sprinkler_info=True)  # makes a dataframe from the data so far
        df_packet = df.drop_duplicates(subset=['raw_packet'])  # drops duplicates of packets
        self.group_tbp = df_packet.groupby(by=['attenuation', 'adv_address']).agg(
            tbp_mean=('tbp', lambda x: np.ceil(np.mean(x[x > 0]))),  # Calculates mean, then applies ceiling
            tbp_min=('tbp', lambda x: np.ceil(np.min(x[x > 0]))),
            tbp_median=('tbp', lambda x: np.ceil(np.median(x[x > 0])))  # Calculates min, then applies ceiling
        ).reset_index()

        self.group_tbp = self.group_tbp.sort_values(by=['adv_address', 'attenuation'])

        self.group_tbp['Difference'] = self.group_tbp.groupby('adv_address')['tbp_min'].diff().fillna(0)
        self.group_tbp = self.group_tbp.dropna()

        # Various calculations on the TBP packets
        self.group_tbp.to_csv(os.path.join(self.results_path, "tbp_calculated_{}.csv".format(run_start_time)))
        # Saves packets with TBP only
        df_packet.to_csv(os.path.join(self.results_path, "filtered_tbp_packets_{}.csv".format(run_start_time)))
        # Saves all the results
        df.to_csv(os.path.join(self.results_path, "full_Test_results_{}.csv".format(run_start_time)))

    def analysis_plot(self):
        self.optimal_value_calculation()
        for adva in self.group_tbp['adv_address'].unique():
            group_data = self.group_tbp[self.group_tbp['adv_address'] == adva]
            y1 = self.group_tbp['tbp_min']
            y2 = self.group_tbp['tbp_mean']
            y3 = self.group_tbp['tbp_median']
            plt.plot(group_data['attenuation'], y1, marker='o', label="tbp_min")

            # Used for debugging and comparing results
            # plt.plot(group_data['attenuation'], y2, marker='o', label="tbp_mean")
            # plt.plot(group_data['attenuation'], y3, marker='o', label="tbp_median")

        plt.draw()
        plt.xlabel("Attenuation")
        plt.ylabel('TBP')
        if self.recommended_stabilization:
            plt.title(f"Recommended attenuation: {self.recommended_stabilization}")
        else:
            plt.title("No recommended attenuation based on criteria")
        plt.legend()
        # TODO: add recommended attenuation to graph
        plt.savefig((os.path.join(self.results_path, "beacon_calibration_plot_{}.png".format(run_start_time))))
        plt.show()

    def optimal_value_calculation(self):
        self.group_tbp['Incremental_Difference'] = self.group_tbp['tbp_min'].diff()

        stabilization_range = 1  # Max differance between points that fit the criteria
        min_consecutive_points = 5  # How many points in a row should be in order to recommend a value

        streak_count = 0  # Tracks the longest streak of values within the stabilization range
        stabilization_start_index = None
        try:
            for index, row in self.group_tbp.iterrows():
                if abs(row['Incremental_Difference']) <= stabilization_range:
                    streak_count += 1
                    if streak_count == min_consecutive_points:
                        stabilization_start_index = index - min_consecutive_points + 1
                        stabilization_start_value = self.group_tbp.loc[stabilization_start_index, 'attenuation']
                        break
                else:
                    streak_count = 0  # Resets said counter if conditions are not met

        except KeyError:
            pass

        if stabilization_start_index is not None:
            self.logger.critical(
                f"Stabilization starts at Attenuation: {stabilization_start_index}, tbp_min: {int(self.group_tbp.loc[stabilization_start_index, 'tbp_min'])}")
            self.recommended_stabilization = stabilization_start_value + OPTIMAL_ATTENUATION_MODIFIER
            messagebox.showinfo(message=f"Recommended Attenuation: {int(stabilization_start_value) + 2}")

        else:
            print("No stabilization detected based on the criteria :(")

    # GUI
    def init_gui(self):

        json_params_path = os.path.join(os.path.dirname(__file__), 'beacon_power_calibration_params.json')

        def load_params_from_json(json_params_path):
            if os.path.exists(json_params_path):
                try:
                    with open(json_params_path, 'r') as json_file:
                        return json.load(json_file)
                except json.JSONDecodeError:
                    print("Error decoding JSON from file")
            return {}

        used_values = load_params_from_json(json_params_path)

        # TODO
        # add the attenuation ports to the GUI

        params_dict = {
            'time_profile_on': {
                'text': 'On',
                'value': used_values.get('time_profile_on', '5'),
                'widget_type': 'entry',
                'group': 'Time profile'
            },
            'time_profile_period': {
                'text': 'Period',
                'value': used_values.get('time_profile_period', '15'),
                'widget_type': 'entry',
                'group': 'Time profile'
            },
            'lora_value': {
                'text': 'LoRa Value',
                'value': used_values.get('lora_value', '20'),
                'widget_type': 'entry',
                'options': {'font': ('Helvetica', 10, 'bold')},
            },
            'iter_length': {
                'text': 'Iteration length (in seconds)',
                'value': used_values.get('iter_length', '90'),
                'widget_type': 'entry',
                'options': {'font': ('Helvetica', 10, 'bold')},
            },
            'energy_pattern_value': {
                'text': 'Energy pattern',
                'value': used_values.get('energy_pattern_value', '50'),
                'widget_type': 'entry',
                'options': {'font': ('Helvetica', 10, 'bold')},
            },
            'sweep_row': [
                {'ble_sweep_start': {
                    'text': 'BLE Sweep Start',
                    'value': '0',
                    'widget_type': 'spinbox',
                    'options': list(range(0, 31)),
                },
                },
                {'ble_sweep_end': {
                    'text': 'to',
                    'value': '30',
                    'widget_type': 'spinbox',
                    'options': list(range(0, 31)),
                },
                }
            ],

            'LORA_ATTN_COM': {
                'text': 'LoRa COM Port',
                'value': used_values.get('LORA_ATTN_COM', 'COM3'),
                'widget_type': 'entry',
                'options': {'font': ('Helvetica', 10, 'bold')},
                'group': 'COM Port Configuration'
            },
            'BLE_ATTN_COM': {
                'text': 'BLE COM Port',
                'value': used_values.get('BLE_ATTN_COM', 'COM6'),
                'widget_type': 'entry',
                'options': {'font': ('Helvetica', 10, 'bold')},
                'group': 'COM Port Configuration'
            },
        }

        beacon_power_calib_gui = WiliotGui(params_dict=params_dict, title='Beacon Power Calibration')
        values = beacon_power_calib_gui.run()
        print(values)
        try:
            form_values = {
                'iter_duration': int(values['iter_length']),
                'lora_value': int(values['lora_value']),
                'sweep_row_ble_sweep_start': values['sweep_row_ble_sweep_start'],
                'sweep_row_ble_sweep_end': values['sweep_row_ble_sweep_end'],
                'energy_pattern_value': int(values['energy_pattern_value']),
                'time_profile_on': int(values['time_profile_on']),
                'time_profile_period': int(values['time_profile_period']),
                'LORA_ATTN_COM': values['LORA_ATTN_COM'],
                'BLE_ATTN_COM': values['BLE_ATTN_COM'],
            }

            # Serialize form values to JSON and write to a file
            with open('beacon_power_calibration_params.json', 'w') as json_file:
                json.dump(form_values, json_file, indent=4)

        except ValueError:
            popup_message('Value must be an integer.', logger=self.logger)


if __name__ == '__main__':
    inst = BeaconCalibration()
    SystemExit("We're done here, get on with your day mate")



"""
  Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
import logging
import time

from wiliot_testers.offline.configs.global_vars_and_enums import TagTestingDefaults
from wiliot_core import WiliotGateway


class TadbikR2rController():
    def __init__(self, gw_name=None):
        """Initialize the controller and its components."""
        self.cleanup_done = False
        self.GwObj = WiliotGateway(auto_connect=True, logger_name='root', device_name=gw_name)
        self.configure_gateway()
        logging.info("tadbikR2rController initialized.")

    def configure_gateway(self):
        """Configure the gateway with initial settings."""
        try:
            self.GwObj.reset_gw()
            if not self.GwObj.is_gw_alive():
                time.sleep(5)
            self.GwObj.config_gw(energy_pattern_val=18, time_profile_val=[0, 15],
                                 pl_delay_val=0, start_gw_app=False, with_ack=True)
            self.GwObj.write('!pl_gw_config 1')  # Enable production line trigger
        except Exception as e:
            logging.error(f"Error configuring gateway: {e}")
            raise

    def __enter__(self):
        """Enter method for context management."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Cleanup resources on exit."""
        if not self.cleanup_done:
            self.cleanup()
        if exc_type:
            logging.error(f"Exception occurred: {exc_value}")
        return False  # To propagate exceptions, if any

    def cleanup(self):
        """Clean up resources."""
        self.cleanup_done = True
        self.stop()
        self.GwObj.reset_gw()
        time.sleep(2)
        self.GwObj.exit_gw_api()
        self.cleanup_done = True
        logging.info("Resources cleaned up.")

    def stop(self):
        self.GwObj.gpio_state(gpio=TagTestingDefaults.START_STOP_GPIO, state='OFF')

    def start(self):
        self.GwObj.gpio_state(gpio=TagTestingDefaults.START_STOP_GPIO, state="ON")

    def failed(self):
        self.GwObj.pulse(gpio=TagTestingDefaults.FAIL_PULSE_GPIO, pulse_duration_ms=50)

    def passed(self):
        self.GwObj.pulse(gpio=TagTestingDefaults.PASS_PULSE_GPIO, pulse_duration_ms=50)

    def set_missing_label_state(self, state="ON"):
        self.GwObj.gpio_state(gpio=TagTestingDefaults.MISSING_LABEL_GPIO, state=state)

    def toggle_direction(self, direction='backward'):
        state = 'ON' if direction == 'backward' else 'OFF'
        self.GwObj.gpio_state(gpio=TagTestingDefaults.PRINTER_DIRECTION_GPIO, state=state)

    def restart(self):
        self.stop()
        self.start()

    def is_r2r_moved(self):
        gw_answer = self.GwObj.read_specific_message(msg="Start Production Line GW", read_timeout=1)
        if gw_answer == '':
            return False
        else:
            self.GwObj.write('!cancel', with_ack=True)
            return True

    def rewind(self, max_missing_labels=6, num_tags=0):
        """
        Rewind the reel with given parameters.

        :param max_missing_labels: Maximum number of missing labels before stopping the rewind.
        :param num_tags: The number of tags to rewind through.
        :raises ValueError: If max_missing_labels is not positive.
        """
        if max_missing_labels <= 0:
            raise ValueError(f"max missing labels must be bigger than zero, got {max_missing_labels}")

        self._start_rewind_process()
        try:
            missing_labels_stop = self._process_rewind_loop(max_missing_labels, num_tags)
            if missing_labels_stop:
                self._handle_missing_labels_stop()
        finally:
            self.cleanup()

    def _start_rewind_process(self):
        """Initialize the rewind process."""
        self.GwObj.start_continuous_listener()
        self.restart()
        self.toggle_direction()

    def _process_rewind_loop(self, max_missing_labels, num_tags):
        """Process the main rewind loop."""
        missing_label_counter, tags_counter = 0, 0
        while True:
            self.passed()
            if not self.is_r2r_moved():
                missing_label_counter += 1
                print(f"Missing labels: {missing_label_counter}")
                if missing_label_counter >= max_missing_labels:
                    print(f"Rewind finished after {missing_label_counter} missing labels")
                    return True  # Indicates missing labels stop condition
            else:
                tags_counter += 1
                missing_label_counter = 0
                if num_tags and tags_counter >= num_tags:
                    print(f"Rewind finished after {tags_counter} tags")
                    break
                if tags_counter % 100 == 0:
                    print(f'Rewinding {tags_counter} tags')

    def _handle_missing_labels_stop(self):
        """Handle additional steps if the missing labels stop condition was met."""
        print("Start searching for first tag")
        self.stop()
        self.toggle_direction('forward')
        self.start()
        for _ in range(100):
            if self.is_r2r_moved():
                print("Roll to first tag")
                return
            self.restart()
            self.passed()
        print("Start of reel wasn't found for 100 tags!")


if __name__ == '__main__':
    tadbik_r2r_controller = TadbikR2rController()
    tadbik_r2r_controller.rewind(max_missing_labels=10, num_tags=1000)
    print("Done!")



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""


import json
import logging
from os.path import join, abspath
import os
import time
from wiliot_api import ManufacturingClient
from wiliot_core import WiliotGateway, GetApiKey
from wiliot_testers.utils.upload_to_cloud_api import upload_to_cloud_api
from wiliot_testers.utils.get_version import get_version
from wiliot_testers.wiliot_tester_log import WiliotTesterLog, dict_to_csv
from wiliot_testers.wiliot_tester_tag_test import WiliotTesterTagTest
from wiliot_testers.wiliot_tester_tag_result import FailureCodes


class SimpleTest:

    def __init__(self,
                 owner_id='wiliot-ops',
                 output_path=None,
                 test_name=None,
                 test_time=None,
                 tester_type='sample',
                 test_env='test',
                 upload_to_cloud=False):
        """
                Initialize the SampleChamber object.

                Parameters: - mode (str): The mode of operation, either "online" or "offline". - external_id (str):
                The external ID of the tag. - test (str): The name of the test to perform according to the tests in
                .default_test_configs.json. - env (str): The environment for cloud connection, either "prod" or
                "test". - surface (str): The surface to simulate, default is 'no simulation' - data according to the
                .default_surfaces.json. - output_path (str): The path where output files will be stored,
                if not mentioned - it will be saved in <USER>/AppData/Local/wiliot/sample/logs directory.

                Initializes various instance variables and starts the test.
                """

        self.run_data_path = None
        self.packets_data_path = None
        self.name = test_name if test_name is not None else 'SimpleTest'
        self.output = output_path
        self.test_data = None
        self.runDataDict = None
        self.gw_version = None
        self.client = None
        self.tester_type = tester_type
        cur_time = time.strftime("_%d%m%y_%H%M%S")
        self.common_run_name = self.name + cur_time
        self.pywiliot_version = get_version()
        self.test = 'Simple Test'
        self.test_suite = None
        self.env = test_env
        self.owner_id = owner_id
        self.packets_csv = None

        self.logger = self.test_configs(test_time=test_time)  # Read all default values from JSONs
        self.gateway_obj = self.com_connect()  # Connect to all HW
        # Init Wiliot Tester SDK
        self.wiliot_tag_test = WiliotTesterTagTest(test_suite=self.test_suite,
                                                   selected_test=self.test,
                                                   logger_name=self.logger.logger.name,
                                                   logger_result_name=self.logger.results_logger.name,
                                                   logger_gw_name=self.logger.gw_logger.name,
                                                   gw_obj=self.gateway_obj)

        self.connect_to_cloud()

        result = self.start_test()

        if result and upload_to_cloud:
            self.logger.logger.info(f'Test done, \nData stored at {self.logger.log_path} \nStarting uploading to cloud')
            upload = self.cloud_upload()

            if not upload:
                self.logger.logger.warning('Failed to upload data to the cloud, please upload it manually')

        self.close_gw_port()

    def test_configs(self, test_time=None):
        """
        Configure the test environment.

        Reads configurations from JSON files and validates them.
        Also sets up logging and test suites according to the test provided.
        """

        logger = None
        try:

            logger = WiliotTesterLog()
            logger.set_common_run_name(reel_name='SimpleTest')
            logger.set_logger(log_path=self.output, tester_name='SimpleTest')
            self.run_data_path, self.packets_data_path = logger.create_data_dir(self.output,
                                                                                tester_name='SimpleTest',
                                                                                run_name=self.common_run_name)
            logger.set_console_handler_level(logger.logger, logging.DEBUG)
            logger.set_console_handler_level(logger.gw_logger, logging.WARNING)
            logger.set_console_handler_level(logger.results_logger, logging.DEBUG)
            sample_test_dir = os.getcwd()
            config_dir = abspath(join(sample_test_dir, 'configs'))

            with open(abspath(join(config_dir, 'test_suite_example.json')), 'r') as TestSuite:
                self.test_suite = json.load(TestSuite)
                if test_time is not None:
                    self.test_suite[self.test]['maxTtfp'] = int(test_time)
                    self.test_suite[self.test]['tests'][0]['maxTime'] = int(test_time)

        except Exception as e:
            if logger is not None:
                logger.logger.error(f'Could not read test suite: {e}')
            raise Exception(e)

        return logger

    def com_connect(self):
        """
                Connect to all hardware components.

                Parameters:
                - default_coms (dict): Dictionary containing default COM port configurations.

                Connects to the Gateway, Attenuators [BLE and LORA],
                Barcode Scanner (if available), and Temperature Sensors (if available).
                """
        gateway_obj = None
        try:
            # Connect to the GW
            gateway_obj = WiliotGateway(auto_connect=True,
                                        logger_name=self.logger.gw_logger.name)
            if gateway_obj.is_connected():
                self.gw_version = gateway_obj.get_gw_version()
                self.logger.logger.info('GW connected')

            else:
                raise Exception('GW is not connected')

        except Exception as e:
            self.logger.logger.error(
                f'Could not detect GW, {e}')
            exit()

        return gateway_obj

    def connect_to_cloud(self):
        """
        Establishes a secure connection to the cloud service for data storage and validation.

        Parameters: - env (str): Specifies the cloud environment to connect to. It can be either 'prod' for
        production or 'test' for testing.

        Steps:
        1. Reads the user's cloud configuration file to obtain API keys and owner IDs.
        2. Validates the owner ID against a predefined list of valid owner IDs.
        3. Initializes the cloud client using the API key and specified environment.
        4. Logs the successful establishment of the cloud connection.

        Note: If the connection fails or configurations are invalid, the function will raise an exception and close
        the Gateway port.
        """
        try:
            g = GetApiKey(gui_type='ttk', env=self.env, owner_id=self.owner_id)
            api_key = g.get_api_key()
            if not api_key:
                raise Exception(f'Could not found an api key for owner id {self.owner_id} and env {self.env}')

            self.client = ManufacturingClient(api_key=api_key, env=self.env, logger_=self.logger.logger.name)
            self.logger.logger.info('Connection to the cloud was established')

        except Exception as e:
            self.close_gw_port()
            raise Exception(f'Problem connecting to cloud {e}')

    def close_gw_port(self):
        """
                Close the connection to the Gateway.

                If the Gateway is connected, this function will close the port.
                """
        self.gateway_obj.close_port()
        self.gateway_obj.exit_gw_api()

    def start_test(self):
        """
        Initiates and executes the test sequence for Wiliot pixie.

        This function performs several key tasks:
        1. Initializes the data structures for storing test results in CSV format.
        2. Executes the test by calling the `run()` method on the `wiliot_tag_test` object.
        3. Collects statistics and summary information about the test.
        4. Populates the run data dictionary with test results and statistics.
        5. Updates the packets data CSV with packet-level information.
        6. If in 'online' mode, it also validates the test results by comparing them with cloud data.

        Returns:
        - True: If the test and optional validation are successful.
        - False: If the test or validation fails.

        Note: This function will also close the Gateway port in case of an exception.
        """
        try:
            self.initialize_csv_data()  # Init the run result dictionary
            res = self.wiliot_tag_test.run()  # Begin the run

            if res.is_results_empty():
                #   Failed
                self.logger.logger.error('Test failed')
                if self.tester_type == 'sample':
                    self.runDataDict['tested'] = 0
                else:
                    self.runDataDict['total_run_tested'] = 0
                self.runDataDict['responded'] = 0
                self.runDataDict['passed'] = 0
                raise Exception('something went wrong during test and test was not preformed')

            selected_tag_statistics = res.tests[0].selected_tag.get_statistics()  # Get all the statistics
            test_sum = res.tests[0].get_summary()

            if self.tester_type == 'sample':
                self.runDataDict['tested'] = 1
            else:
                self.runDataDict['total_run_tested'] = 1
            self.runDataDict['testTime'] = str(res.get_total_test_duration())
            self.runDataDict['passed'] = int(res.is_all_tests_passed())
            self.runDataDict['responded'] = int(1 if res.tests[0].all_packets.__len__() > 0 else 0)
            if self.tester_type == 'sample':
                self.runDataDict['responding[%]'] = str(
                    self.runDataDict['responded'] / self.runDataDict['tested'] * 100) + '%'
                self.runDataDict['passed[%]'] = str(self.runDataDict['passed'] / self.runDataDict['tested'] * 100) + '%'
            else:
                self.runDataDict['responding[%]'] = str(
                    self.runDataDict['responded'] / self.runDataDict['total_run_tested'] * 100) + '%'
                self.runDataDict['passed[%]'] = str(
                    self.runDataDict['passed'] / self.runDataDict['total_run_tested'] * 100) + '%'

            if self.runDataDict['passed']:
                self.logger.logger.info('Test passed')
            elif self.runDataDict['responded']:
                selected_tag_statistics = res.tests[0].all_packets.get_statistics()
                self.logger.logger.info('Test failed, tag responded but tbp could not be calculated')
            else:
                self.logger.logger.info('Test failed, tag did not respond')

            # save files:
            self.populate_run_data(selected_tag_statistics, test_sum)  # Fill run result dictionary
            self.update_packets_data(res, selected_tag_statistics)

            self.wiliot_tag_test.exit_tag_test()
            return True

        except Exception as e:
            self.close_gw_port()
            self.logger.logger.error(f'Problem starting the test {e}')
            return False

    def initialize_csv_data(self):
        """
                Initialize the CSV data structures.

                Sets up dictionaries for run data and packet data.
                """

        test_values = self.test_suite[self.test]['tests'][0]
        if self.tester_type == 'sample':
            self.runDataDict = {
                'testerStationName': 'WiliotSample',
                'commonRunName': self.common_run_name,
                'batchName': '',
                'testerType': self.tester_type,
                'comments': '',
                'errors': '',
                'timeProfile': test_values.get('timeProfile', ''),
                'txPower': test_values.get('absGwTxPowerIndex', ''),
                'energizingPattern': test_values.get('energizingPattern', ''),
                'tested': '0',
                'passed': '0',
                'yield': '0',
                'inlay': '',
                'responded': '0',
                'responding[%]': '0',
                'passed[%]': '0',
                'testStatus': 'False',
                'operator': 'Wiliot',
                'testTime': '',
                'runStartTime': '',
                'runEndTime': '',
                'antennaType': 'TIKI',
                'surface': '',
                'numChambers': '1',
                'gwVersion': self.gw_version,
                'pyWiliotVersion': self.pywiliot_version,
                'bleAttenuation': '',
                'loraAttenuation': '',
                'testTimeProfilePeriod': test_values.get('timeProfile'[1], ''),
                'testTimeProfileOnTime': test_values.get('timeProfile'[0], ''),
                'ttfpAvg': '',
                'tbpAvg': '',
                'tbpStd': '',
                'rssiAvg': '',
                'maxTtfp': '',
                'controlLimits': '',
                'hwVersion': '',
                'sub1gFrequency': '',
                'failBinStr': '',
                'failBin': ''
            }

            self.test_data = {
                'commonRunName': self.common_run_name,
                'encryptedPacket': '',
                'time': '',
                'reel': self.common_run_name,
                'ttfp': '',
                'tbp': '',
                'adv_address': '',
                'status': ''}
        else:
            self.runDataDict = {
                'testerStationName': 'WiliotSample',
                'commonRunName': self.common_run_name,
                'batchName': '',
                'testerType': self.tester_type,
                'comments': '',
                'errors': '',
                'timeProfile': test_values.get('timeProfile', ''),
                'txPower': test_values.get('absGwTxPowerIndex', ''),
                'energizingPattern': test_values.get('energizingPattern', ''),
                'total_run_tested': '0',
                'passed': '0',
                'yield': '0',
                'inlay': '',
                'responded': '0',
                'responding[%]': '0',
                'passed[%]': '0',
                'testStatus': 'False',
                'operator': 'Wiliot',
                'testTime': '',
                'runStartTime': '',
                'runEndTime': '',
                'antennaType': 'TIKI',
                'surface': '',
                'numChambers': '1',
                'gwVersion': self.gw_version,
                'pyWiliotVersion': self.pywiliot_version,
                'bleAttenuation': '',
                'loraAttenuation': '',
                'testTimeProfilePeriod': test_values.get('timeProfile'[1], ''),
                'testTimeProfileOnTime': test_values.get('timeProfile'[0], ''),
                'ttfpAvg': '',
                'test_suite_dict': self.test_suite[self.test],
                'owner_id': self.owner_id,
                'to_print': False,
                'qr_validation': False,
                'tbpAvg': '',
                'tbpStd': '',
                'rssiAvg': '',
                'maxTtfp': '',
                'controlLimits': '',
                'hwVersion': '',
                'sub1gFrequency': '',
                'failBinStr': '',
                'failBin': ''
            }

            self.test_data = {
                'common_run_name': self.common_run_name,
                'raw_packet': '',
                'adv_address': '',
                'time_from_start': '',
                'ttfp': '',
                'tbp_mean': '',
                'rssi': '',
                'gw_packet': '',
                'tag_run_location': '0',
                'external_id': '',
                'status_offline': '',
                'selected_tag': '',
                'fail_bin': '',
                'fail_bin_str': '',
                'test_num': 0,
                'is_test_pass': '',
                'packet_status': ''}

        dict_to_csv(dict_in=self.runDataDict, path=self.run_data_path, append=False, only_titles=True)
        dict_to_csv(dict_in=self.test_data, path=self.packets_data_path, append=False, only_titles=True)

    def populate_run_data(self, selected_tag_statistics, test_sum):
        """
                Populate the run data dictionary.

                Parameters:
                - selected_tag_statistics (dict): Statistics of the selected tag.
                - test_sum (dict): Summary of the test.

                Fills the run data dictionary with test results and statistics.
                """
        self.runDataDict['testStatus'] = bool(test_sum['is_test_pass'])
        self.runDataDict['runStartTime'] = test_sum['test_start_time'].strftime("%Y-%m-%d %H:%M:%S.%f")
        self.runDataDict['runEndTime'] = test_sum['test_end_time'].strftime("%Y-%m-%d %H:%M:%S.%f")
        self.runDataDict['ttfpAvg'] = str(selected_tag_statistics.get('ttfp', 'nan'))
        self.runDataDict['tbpAvg'] = str(selected_tag_statistics.get('tbp_mean', 'nan'))
        self.runDataDict['tbpStd'] = str(selected_tag_statistics.get('tbp_std', 'nan'))
        self.runDataDict['rssiAvg'] = str(selected_tag_statistics.get('rssi_mean', 'nan'))
        self.runDataDict['maxTtfp'] = str(selected_tag_statistics.get('ttfp', 'nan'))
        dict_to_csv(dict_in=self.runDataDict, path=self.run_data_path, append=True, only_titles=False)

    def update_packets_data(self, res=None, stats=None):
        """
                Update the packets data CSV.

                Parameters:
                - res (object): The result object containing test data.
                - stats (dict): Statistics of the selected tag.

                Updates the packets data CSV with packet information.
                """

        if res:
            r = res.tests[0]
            if r.selected_tag.__len__() > 0:  # pass
                tag_data = r.selected_tag
            elif r.all_packets.__len__() > 0:  # responded but did not pass
                tag_data = r.all_packets
            else:  # No response
                if self.tester_type == 'sample':
                    self.test_data.update({
                        'encryptedPacket': '',
                        'time': '',
                        'status': r.is_test_passed,
                        'ttfp': str(stats.get('ttfp', 'nan')),
                        'tbp': str(stats.get('tbp_min', 'nan')),
                        'adv_address': ''})
                else:
                    self.test_data.update({
                        'raw_packet': '',
                        'time_from_start': '',
                        'packet_status': r.is_test_passed,
                        'ttfp': str(stats.get('ttfp', 'nan')),
                        'tbp_mean': str(stats.get('tbp_min', 'nan')),
                        'rssi': str(stats.get('rssi', 'nan')),
                        'tag_run_location': '',
                        'status_offline': 0,
                        'gw_packet': '',
                        'fail_bin': FailureCodes.NO_RESPONSE.value,
                        'fail_bin_str': FailureCodes.NO_RESPONSE.name,
                        'is_test_pass': r.is_test_passed,
                        'adv_address': ''})
                return

            if self.tester_type == 'sample':
                for p in tag_data:
                    self.test_data.update({
                        'encryptedPacket': p.get_packet(),
                        'time': p.gw_data['time_from_start'][0] if p.gw_data['time_from_start'].size > 1 else p.gw_data[
                            'time_from_start'].item(),
                        'status': r.is_test_passed,
                        'ttfp': str(stats.get('ttfp', 'nan')),
                        'tbp': str(stats.get('tbp_min', 'nan')),
                        'adv_address': p.packet_data['adv_address']})
                    dict_to_csv(dict_in=self.test_data, path=self.packets_data_path, append=True, only_titles=False)
            else:
                for p in tag_data:
                    for i in range(p.gw_data['gw_packet'].size):
                        if p.gw_data["gw_packet"].size > 1:
                            gw_packet = p.gw_data["gw_packet"][i]
                            rssi = p.gw_data["rssi"][i]
                            time_from_start = p.gw_data["time_from_start"][i]
                        else:
                            gw_packet = p.gw_data["gw_packet"].item()
                            rssi = p.gw_data["rssi"].item()
                            time_from_start = p.gw_data["time_from_start"].item()
                        self.test_data.update({
                            'raw_packet': p.get_packet(),
                            'status_offline': 1,
                            'time_from_start': time_from_start,
                            'packet_status': r.is_test_passed,
                            'ttfp': str(stats.get('ttfp', 'nan')),
                            'tbp_mean': str(stats.get('tbp_min', 'nan')),
                            'rssi': rssi,
                            'gw_packet': gw_packet,
                            'fail_bin': FailureCodes.PASS.value,
                            'fail_bin_str': FailureCodes.PASS.name,
                            'is_test_pass': r.is_test_passed,
                            'selected_tag': p.packet_data['adv_address'],
                            'adv_address': p.packet_data['adv_address']})
                        dict_to_csv(dict_in=self.test_data, path=self.packets_data_path, append=True, only_titles=False)

    def cloud_upload(self):
        """
        Uploads the collected test data to the cloud for further analysis and storage.

        Parameters: - env (str): Specifies the cloud environment to upload to. It can be either 'prod' for production
        or 'test' for testing.

        The function performs the following tasks:
        1. Calls the `upload_to_cloud_api` function to upload both run data and packets data CSV files.
        2. Logs the status of the upload operation.

        Returns:
        - True: If the data upload to the cloud is successful.
        - False: If the data upload fails for any reason.

        Note: The function uses the owner ID and API key that were set during the cloud connection phase.
        """
        tester_type = self.tester_type + '-test'
        success = upload_to_cloud_api(batch_name='',
                                                  tester_type=tester_type,
                                                  run_data_csv_name=self.run_data_path,
                                                  packets_data_csv_name=self.packets_data_path,
                                                  is_path=True,
                                                  env=self.env,
                                                  owner_id=self.owner_id,
                                                  logger_=self.logger.logger.name)

        return success

    def get_common_run_name(self):
        return self.common_run_name

    def get_out_put_dir(self):
        return self.output


if __name__ == "__main__":
    Test = SimpleTest()
